// $Id: ec.cc,v 1.3 1999/05/04 19:21:27 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/ec.cc,v $
//
// $Log: ec.cc,v $
// Revision 1.3  1999/05/04 19:21:27  mencnar
// Fixes in ORBIX port.
//
// Revision 1.2  1999/02/24 12:27:59  mencnar
// added keep-alive calls
//
// Revision 1.1.1.1  1998/12/13 23:27:15  mencda
// Imported sources
//
//

//
// Main program module
// Implements event channels and channel proxies, except for multicast forwarding
// 

#ifdef ORBIX
#define EXCEPTIONS
#endif

#include "ec.hh"
#include "bootstrap.hh"
#include "IRread.hh"
#include "trace.hh"

//
// an object responsible for giving out unique identifier for channels
// and consumers
//
UniqueID<CORBA::ULong> EventChannel_i::channel_id_dispenser( 0);

//------------------------------------------------------------EventChannel_i--
//
EventChannel_i::
EventChannel_i( CORBA::ULong _max_queue,
		ECAdmin::OverflowAction _overflow_action,
		CORBA::ULong _not_connected_timeout,
		CORBA::ULong _keep_alive_time,
		AbstractForwarderInstance _forwarder,
		MyECController *_eccontr) :
  checker_cond( &checker_guard),
  consumer_id_dispenser(0),
  eccontr( _eccontr),
  consumer_admin( new ConsumerAdmin_i(*this)),
  supplier_admin( new SupplierAdmin_i(*this)),
  is_typed( 0),
  forwarder( _forwarder),
  channel_id( channel_id_dispenser.assign_id()),
  default_max_queue( _max_queue),
  default_overflow_action( _overflow_action),
  not_connected_timeout( _not_connected_timeout),
  keep_alive_time( _keep_alive_time),
  is_active( 1)
{
  corba_object= new SK_EventChannel( *this);
  CONNECT_IMPL(Bootstrap::boa_ptr, corba_object);
  if (*forwarder) forwarder->open_event_channel( *this, 
						 forwarder,
						 channel_id,
						 default_max_queue,
						 default_overflow_action,
						 keep_alive_time);

  checker_ec_instance= new ECRef( *this);
  (new omni_thread( static_checker_worker, this))->start();

  trace.start(50) << "Event_Channel_i ctor" << '\n';
}

EventChannel_i::
EventChannel_i( CORBA::ULong _max_queue,
		ECAdmin::OverflowAction _overflow_action,
		CORBA::ULong _not_connected_timeout,
		CORBA::ULong _keep_alive_time,
		AbstractForwarderInstance _forwarder,
		IRMOD InterfaceDef_ptr _idef,
		MyECController *_eccontr) :
  checker_cond( &checker_guard),
  consumer_id_dispenser(0),
  eccontr( _eccontr),
  is_typed( 1),
  consumer_admin( new ConsumerAdmin_i(*this)),
  supplier_admin( new SupplierAdmin_i(*this)),
  forwarder( _forwarder),
  channel_id( channel_id_dispenser.assign_id()),
  default_max_queue( _max_queue),
  default_overflow_action( _overflow_action),
  not_connected_timeout( _not_connected_timeout),
  keep_alive_time( _keep_alive_time),
  idef( IRMOD InterfaceDef::_duplicate( _idef)),
  is_active( 1)
{
  try {
    if (!check_interface( idef, tpintf)) {
      trace.start(50) << "Interface "  
		      << " does not fulfil contraints for typed event channel"
		      << '\n';
      throw 1;
    }
  }
  catch (CORBA::Exception &ex) {
    //    trace.start(50) << ex << '\n';
    throw 1;
  }
  typed_corba_object= new SK_TypedEventChannel( *this);
  CONNECT_IMPL(Bootstrap::boa_ptr, typed_corba_object);
  if (*forwarder) forwarder->open_event_channel( *this, 
						 forwarder,
						 channel_id,
						 default_max_queue,
						 default_overflow_action,
						 keep_alive_time);
  checker_ec_instance= new ECRef( *this);
  (new omni_thread( static_checker_worker, this))->start();

  trace.start(50) << "Event_Channel_i ctor" << '\n';
}

EventChannel_i::
~EventChannel_i() {
  trace.start(50) << "~EventChannel_i" << '\n';
}

void EventChannel_i::
check_interface_name( const char *intf) {
  // NOT IMPLEMENTED
}

CosEventChannelAdmin::ConsumerAdmin_ptr EventChannel_i::
for_consumers() {

  return CosEventChannelAdmin::ConsumerAdmin::
    _duplicate( consumer_admin->corba_object);
}

CosEventChannelAdmin::SupplierAdmin_ptr EventChannel_i::
for_suppliers() {
  
  return CosEventChannelAdmin::SupplierAdmin::
    _duplicate( supplier_admin->corba_object);
}

CosTypedEventChannelAdmin::TypedConsumerAdmin_ptr EventChannel_i::
for_typed_consumers() {

  return CosTypedEventChannelAdmin::TypedConsumerAdmin::
    _duplicate( consumer_admin->typed_corba_object);
}

CosTypedEventChannelAdmin::TypedSupplierAdmin_ptr EventChannel_i::
for_typed_suppliers() {

  trace.start(100) << "for_typed_suppliers" << '\n';
  return CosTypedEventChannelAdmin::TypedSupplierAdmin::
    _duplicate( supplier_admin->typed_corba_object);
}

void EventChannel_i::
static_checker_worker( void *args) {
  ((EventChannel_i*)args)->checker_worker();
  delete ((EventChannel_i*)args)->checker_ec_instance;
}

//
// thread worker:
// - keeps track of unconnected proxies and perform keep-alive calls
//   for all connected clients (except for those handled by a remote daemon)
//
void EventChannel_i::
checker_worker() {
  ECRef ecref( *this);

  TimeStamp wakeup_time(10000L,0L);
  checker_guard.lock();
  int do_sleep= 1;
  for (;;) {
    unsigned long sleep_s, sleep_ns;
    omni_thread::get_time( &sleep_s, &sleep_ns, 
			   wakeup_time.sec, wakeup_time.nsec);
    //    cerr << "Sleep for " << wakeup_time.sec 
    //	 << " " << wakeup_time.nsec << endl;
    if (do_sleep) checker_cond.timedwait( sleep_s, sleep_ns);
    do_sleep= 1;
    //    cerr << "Checker thread active" << endl;

    wakeup_time= TimeStamp(1000000L, 0);

    TimeStamp ct= default_tsf->instantiate();

    for (CheckerProxies::iterator it( checker_proxies.begin());
	 it != checker_proxies.end();) {

      ECProxyState state= (*it)->get_state();
      TimeStamp ts= (*it)->get_timestamp();
      if (state==ECPS_DISCONNECTED) {
	CheckerProxies::iterator it2= it;
	(*it)->release_proxy();
	it++;
	checker_proxies.erase( it2);
	continue;
      }
      else if (state==ECPS_NOT_CONNECTED && not_connected_timeout) {
	//	cerr << "NOT_CONNECTED" << endl;
	TimeStamp timeout( not_connected_timeout, 0);
	TimeStamp df= ct.diff(ts);

	if ( timeout < df) {
	  checker_guard.unlock();
	  //	  cerr << "deactivate proxy" << endl;
	  try {
	    (*it)->deactivate_proxy();
	  }
	  catch (...) {}
	  checker_guard.lock();
	  do_sleep= 0;
	  break;
	}
	else {
	  // update wakeup_time
	  TimeStamp new_df= TimeStamp(not_connected_timeout+1, 0).diff(df);
	  if (new_df < wakeup_time) wakeup_time= new_df;
	}
      }
      else if (state==ECPS_CONNECTED && keep_alive_time) {
	//	cerr << "CONNECTED" << endl;
	TimeStamp timeout( keep_alive_time, 0);
	TimeStamp df= ct.diff(ts);

	if ( timeout < df) {
	  checker_guard.unlock();
	  //	  cerr << "is_alive ";
	  int is_dead= 0;
	  CORBA::Object_var obj= (*it)->get_client_reference();
	  if (!CORBA::is_nil( obj)) {
	    is_dead= 1;
	    try {
	      is_dead= obj->_non_existent();
	    }
	    catch (CORBA::BAD_OPERATION &) {
	      is_dead= 0;
	    }
	    catch (...) {}
	  }
	  //	  cerr << (!is_dead) << endl;
	  if (is_dead) {
	    try {
	      //	      cerr << "deactivate_proxy" << endl;
	      (*it)->deactivate_proxy();
	    }
	    catch (...) {}
	  }
	  (*it)->touch();
	  checker_guard.lock();
	  do_sleep= 0;
	  break;
	}
	else {
	  // update wakeup_time
	  TimeStamp new_df= TimeStamp(keep_alive_time, 0).diff(df);
	  if (new_df < wakeup_time) wakeup_time= new_df;
	}
      }
      // otherwise it is a forwarded proxy or keep-alive calls  has been 
      // tuned off
      //
      it++;
    }
    if (!is_active) {
      // cleanup actions, performed upon channel destruction
      //
      for (CheckerProxies::iterator it( checker_proxies.begin());
	   it != checker_proxies.end();) {
	CheckerProxies::iterator it2= it;
	(*it)->release_proxy();
	it++;
	checker_proxies.erase( it2);
      }
      checker_guard.unlock();
      return;
    }
  }
}

void EventChannel_i::
insert_checker_proxy( ECProxy *proxy) {
  
  omni_mutex_lock oml(checker_guard);
  proxy->duplicate_proxy();
  checker_proxies.push_front(proxy);
  checker_cond.signal();
}

void EventChannel_i::
wakeup_checker() {
  omni_mutex_lock oml(checker_guard);
  checker_cond.signal();
}

void EventChannel_i::
destroy() {
  trace.start(50) << "destroy: initialize event channel shutdown" << '\n';
  ECRef tmp(*this);
  {
    omni_mutex_lock oml(checker_guard);
    is_active= 0;
  }
  wakeup_checker();

  if (is_typed) {
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, typed_corba_object);
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, consumer_admin->typed_corba_object);
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, supplier_admin->typed_corba_object);
  } else {
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, corba_object);
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, consumer_admin->corba_object);
    DEACTIVATE_OBJ( Bootstrap::boa_ptr, supplier_admin->corba_object);
  }
  XSTD::set<ProxyPushConsumer_i*,
            XSTD::less<ProxyPushConsumer_i*> > push_cons;
  XSTD::set<ProxyPushSupplier_i*,
            XSTD::less<ProxyPushSupplier_i*> > push_supp;
  XSTD::set<ProxyPullConsumer_i*,
            XSTD::less<ProxyPullConsumer_i*> > pull_cons;
  XSTD::set<ProxyPullSupplier_i*,
            XSTD::less<ProxyPullSupplier_i*> > pull_supp;
  {
    GeneralLock::auto_read rl(proxy_guard);
    push_cons= proxy_push_consumers;
    push_supp= proxy_push_suppliers;
    pull_cons= proxy_pull_consumers;
    pull_supp= proxy_pull_suppliers;
  }

  for (SetPPushCons::iterator it1= push_cons.begin();
       it1 != push_cons.end();
       it1++) {
    try {
      (*it1)->disconnect_push_consumer();
    }
    catch (...) {}
  }

  for (SetPPushSupp::iterator it2= push_supp.begin();
       it2 != push_supp.end();
       it2++) {
    try {
      (*it2)->disconnect_push_supplier();
    }
    catch (...) {}
  }

  for (SetPPullCons::iterator it3= pull_cons.begin();
       it3 != pull_cons.end();
       it3++) {
    try {
      (*it3)->disconnect_pull_consumer();
    }
    catch (...) {}
  }

  for (SetPPullSupp::iterator it4= pull_supp.begin();
       it4 != pull_supp.end();
       it4++) {
    try {
      (*it4)->disconnect_pull_supplier();
    }
    catch (...) {}
  }
	 
  if (eccontr) eccontr->channel_disposed( this);

  trace.start(50) << "destroy: event channel shutdown initialized" << '\n';  
}

CosEventChannelAdmin::ProxyPushSupplier_ptr EventChannel_i::
obtain_push_supplier() {
  ProxyPushSupplier_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPushSupplier_i(*this, 
				 default_max_queue,
				 default_overflow_action);
    CONNECT_IMPL(Bootstrap::boa_ptr, ret);
    proxy_push_suppliers.insert( ret);
    id_to_proxy_push_supplier[ ret->get_id()]= ret;
    insert_checker_proxy( ret);
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPushSupplier::_duplicate(ret);
}

CosEventChannelAdmin::ProxyPullSupplier_ptr EventChannel_i::
obtain_pull_supplier() {
  ProxyPullSupplier_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPullSupplier_i(*this,
				 default_max_queue,
				 default_overflow_action);
    proxy_pull_suppliers.insert( ret);
    insert_checker_proxy( ret);
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPullSupplier::
    _duplicate( is_typed ? ret->typed_corba_object : ret->corba_object);
}

CosEventChannelAdmin::ProxyPushConsumer_ptr EventChannel_i::
obtain_push_consumer() {
  ProxyPushConsumer_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPushConsumer_i(*this);
    proxy_push_consumers.insert( ret);
    insert_checker_proxy( ret);
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPushConsumer::
    _duplicate( is_typed ? ret->typed_corba_object: ret->corba_object);
}

CosEventChannelAdmin::ProxyPullConsumer_ptr EventChannel_i::
obtain_pull_consumer() {
  trace.start(100) << "obtain_pull_consumer" << '\n';
  ProxyPullConsumer_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPullConsumer_i(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,ret);
    proxy_pull_consumers.insert( ret);
    insert_checker_proxy( ret);
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPullConsumer::_duplicate(ret);
}

CosEventChannelAdmin::ProxyPushSupplier_ptr EventChannel_i::
obtain_typed_push_supplier( const char *intf) {
  trace.start(100) << "obtain_typed_push_supplier" << intf << '\n';
  check_interface_name( intf);

  ProxyPushSupplier_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPushSupplier_i(*this,
				 default_max_queue,
				 default_overflow_action);
    CONNECT_IMPL(Bootstrap::boa_ptr,ret);
    proxy_push_suppliers.insert( ret);
    id_to_proxy_push_supplier[ ret->get_id()]= ret;
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPushSupplier::_duplicate(ret);
}
  
CosTypedEventChannelAdmin::TypedProxyPullSupplier_ptr EventChannel_i::
obtain_typed_pull_supplier( const char *intf) {
  trace.start(100) << "obtain_typed_pull_supplier " << intf << '\n';
  check_interface_name( intf);

  ProxyPullSupplier_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPullSupplier_i(*this,
				 default_max_queue,
				 default_overflow_action);
    proxy_pull_suppliers.insert( ret);
  }
  ret->start();
  return CosTypedEventChannelAdmin::TypedProxyPullSupplier::
    _duplicate( ret->typed_corba_object); // no doubt, it is typed EC
}

CosTypedEventChannelAdmin::TypedProxyPushConsumer_ptr EventChannel_i::
obtain_typed_push_consumer( const char *intf) {
  trace.start(100) << "obtain_typed_push_consumer " << intf << '\n';
  check_interface_name( intf);

  ProxyPushConsumer_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPushConsumer_i(*this);
    proxy_push_consumers.insert( ret);
  }
  ret->start();
  return CosTypedEventChannelAdmin::TypedProxyPushConsumer::
    _duplicate( ret->typed_corba_object); // no doubt, it is typed EC
}

CosEventChannelAdmin::ProxyPullConsumer_ptr EventChannel_i::
obtain_typed_pull_consumer( const char *intf) {
  check_interface_name( intf);

  ProxyPullConsumer_i *ret;
  {
    GeneralLock::auto_write wrl(proxy_guard);
    ret= new ProxyPullConsumer_i(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,ret);
    proxy_pull_consumers.insert( ret);
  }
  ret->start();
  return CosEventChannelAdmin::ProxyPullConsumer::_duplicate(ret);
}

CosEventChannelAdmin::ProxyPushSupplier_ptr EventChannel_i::
find_proxy_push_supplier( CORBA::ULong consumer_id) {
  GeneralLock::auto_read garl(proxy_guard);
  MapIDProxy::iterator it=
    id_to_proxy_push_supplier.find( consumer_id);
  if (it != id_to_proxy_push_supplier.end())
    return CosEventChannelAdmin::ProxyPushSupplier::
      _duplicate( (*it).second);
  else
    return NULL;
}

void EventChannel_i::
register_proxy_push_consumer( ProxyPushConsumer_i &) {
}

void EventChannel_i::
unregister_proxy_push_consumer( ProxyPushConsumer_i &pp) {
  GeneralLock::auto_write wrl(proxy_guard);
  proxy_push_consumers.erase( &pp);
}

void EventChannel_i::
register_proxy_push_supplier( ProxyPushSupplier_i &) {
}

void EventChannel_i::
unregister_proxy_push_supplier( ProxyPushSupplier_i &pp) {
  GeneralLock::auto_write wrl(proxy_guard);
  proxy_push_suppliers.erase( &pp);
  id_to_proxy_push_supplier.erase( pp.get_id());
}


void EventChannel_i::
register_proxy_pull_consumer( ProxyPullConsumer_i &) {
}

void EventChannel_i::
unregister_proxy_pull_consumer( ProxyPullConsumer_i &pp) {
  GeneralLock::auto_write wrl(proxy_guard);
  proxy_pull_consumers.erase( &pp);
}

void EventChannel_i::
register_proxy_pull_supplier( ProxyPullSupplier_i &) {
}

void EventChannel_i::
unregister_proxy_pull_supplier( ProxyPullSupplier_i &pp) {
  GeneralLock::auto_write wrl(proxy_guard);
  proxy_pull_suppliers.erase( &pp);
}

void EventChannel_i::
push( ECDHandler &ecdh) {
  GeneralLock::auto_read rrl(proxy_guard);
  
  if (*forwarder) forwarder->push( ecdh);

  for (SetPPushSupp::iterator it1= 
	 proxy_push_suppliers.begin();
       it1!=proxy_push_suppliers.end();
       it1++)
    (*it1)->push( ecdh);
  
  for (SetPPullSupp::iterator it2= 
	 proxy_pull_suppliers.begin();
       it2!=proxy_pull_suppliers.end();
       it2++)
    (*it2)->push( ecdh);
}

void EventChannel_i::
push( const CORBA::Any &data) {
  ECHelper::ECValue *v= new ECHelper::ECValue();
  v->any_value( data);
  ECDHandler ecdh( new ECData(v), 0);
  push(ecdh);
}

// assume ownership
void EventChannel_i::
push( CORBA::Any *data) {
  ECHelper::ECValue *v= new ECHelper::ECValue();
  v->any_value( *data);
  delete data;
  ECDHandler ecdh( new ECData(v), 0);
  push(ecdh);
}

ConsumerAdmin_i::
ConsumerAdmin_i( EventChannel_i &an_ec_i) :
  ec_i( an_ec_i),
  is_typed( an_ec_i.is_typed)
{
  if (is_typed) {
    typed_corba_object= new SK_TypedConsumerAdmin(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,typed_corba_object);
  }
  else {
    corba_object= new SK_ConsumerAdmin(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,corba_object);
  }
}


SupplierAdmin_i::
SupplierAdmin_i( EventChannel_i &an_ec_i) :
  ec_i( an_ec_i),
  is_typed( an_ec_i.is_typed)
{
  if (is_typed) {
    typed_corba_object= new SK_TypedSupplierAdmin(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,typed_corba_object);
  }
  else {
    corba_object= new SK_SupplierAdmin(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,corba_object);
  }
}

//--------------------------------------------------------ProxyPushConsumer--
// IDL interface : ProxyPushConsumer

ProxyPushConsumer_i::
ProxyPushConsumer_i( EventChannel_i &an_ec_i) : 
  ECProxy( an_ec_i),
  delete_inhibitor( ec_i.is_typed ? 2 : 1)
{
  trace.start(100) << "ProxyPushConsumer_i ctor" << '\n';
  if (ec_i.is_typed) {
    typedimpl= default_ec_dsi_factory->create_ec_dsi( ec_i.tpintf,0,*this);

    typed_corba_object= new SK_TypedProxyPushConsumer( *this);
    CONNECT_IMPL(Bootstrap::boa_ptr,typed_corba_object);
  }
  else {
    corba_object= new SK_ProxyPushConsumer( *this);
    CONNECT_IMPL(Bootstrap::boa_ptr,corba_object);
  }
}

void ProxyPushConsumer_i::
delete_proxy( ProxyPushConsumer_i &ppc) {
  int del_flag= 0;
  {
    omni_mutex_lock oml( ppc.delete_guard);
    if (--ppc.delete_inhibitor == 0) del_flag= 1;
  }
  if (del_flag) delete &ppc;
}

void ProxyPushConsumer_i::
connect_push_supplier( CosEventComm::PushSupplier_ptr push_supp) {
  omni_mutex_lock oml(guard);

  // PRECONDITION:
  if ( state != ECPS_NOT_CONNECTED) 
    throw CosEventChannelAdmin::AlreadyConnected();
  // it is permitted for push_supp to be a nil reference
  ORBIX_NIL_REF_BUG( push_supp);

  // BODY:
  push_supplier= CosEventComm::PushSupplier::_duplicate( push_supp);
  ec_i.register_proxy_push_consumer( *this);
  state= ECPS_CONNECTED;
}

void ProxyPushConsumer_i::
push( const CORBA::Any &data) {

  // PRECONDITION:
  {
    omni_mutex_lock oml(guard);
    if (state != ECPS_CONNECTED) 
      throw CosEventComm::Disconnected();
  }

  // BODY:
  // forward data to the event channel
  ec_i.push( data);
}

void ProxyPushConsumer_i::
disconnect_push_consumer() {
  {
    omni_mutex_lock oml(guard);
    
    // BODY:
    if (state== ECPS_CONNECTED) ec_i.unregister_proxy_push_consumer( *this);
    if (state!= ECPS_DISCONNECTED) {
      state= ECPS_DISCONNECTED;
      
      if (!CORBA::is_nil(push_supplier)) {
	omni_mutex_lock oml2( delete_guard);
	++delete_inhibitor;
	
	(new omni_thread( finalize_worker, this))->start();
      }
      
      DEACTIVATE_OBJ( Bootstrap::boa_ptr, 
		      ec_i.is_typed ? 
		      typed_corba_object : corba_object);
      if (ec_i.is_typed) {
	CORBA::Object_var xobj= typedimpl->get_object();
	DEACTIVATE_OBJ( Bootstrap::boa_ptr, xobj);
      }
    }
  }
  ec_i.wakeup_checker();
}

void ProxyPushConsumer_i::
finalize_worker( void *args) {
  ProxyPushConsumer_i &ppc= *(ProxyPushConsumer_i*) args;
  
  // PRECONDITION: state==ECPS_DISCONNECTED
  //               push_supp is not a NIL reference
  //
  
  try {
    ppc.push_supplier->disconnect_push_supplier();
  }
  catch (...) {}
  delete_proxy( ppc);
}

CORBA::Object_ptr ProxyPushConsumer_i::
get_typed_consumer() {
  return typedimpl->get_object();
}

CORBA::Boolean ProxyPushConsumer_i::
ec_dispatch( CORBA::Object_ptr obj,
	     CORBA::Boolean twoway,
	     const char *operation_name,
	     ECHelper::SeqAny &params) {

  ECHelper::ECValue *ecv= new ECHelper::ECValue();
  ECHelper::TypedOperation tp;
  tp.name= (const char *) operation_name;
  tp.twoway= twoway; // we really 
  tp.params= params;
  ecv->oper(tp);
  ECDHandler ecdh( new ECData(ecv), 0);
  ec_i.push(ecdh);

  return (CORBA::Boolean) 0; // this value should be ignored
}


//--------------------------------------------------------ProxyPushSupplier--
// IDL interface : ProxyPushSupplier

ProxyPushSupplier_i::
ProxyPushSupplier_i( EventChannel_i &an_ec_i,
		     CORBA::ULong _max_queue,
		     ECAdmin::OverflowAction _overflow_action) : 
  ECProxy( an_ec_i),
  thr( NULL),
  queue_not_empty(&guard),
  max_queue( _max_queue),
  overflow_action( _overflow_action),
  forwarded( 0),
  ec_dii( NULL),
  consumer_id( ec_i.consumer_id_dispenser.assign_id())
{
  trace.start(100) << "ProxyPushSupplier_i ctor" << '\n';
}

void ProxyPushSupplier_i::
connect_push_consumer( CosEventComm::PushConsumer_ptr push_cons
		       ENV_DEF(env)) {
  omni_mutex_lock oml(guard);

  // PRECONDITION:
  if ( state != ECPS_NOT_CONNECTED) 
    throw CosEventChannelAdmin::AlreadyConnected();
  // it is not allowed for cons_cons to be nil reference
  ORBIX_NIL_REF_BUG( push_cons);
  if ( CORBA::is_nil(push_cons))
    throw CORBA::BAD_PARAM(0,CORBA::COMPLETED_NO);

  // BODY:
  push_consumer= CosEventComm::PushConsumer::_duplicate( push_cons);
  ec_i.register_proxy_push_supplier( *this);
  if (*(ec_i.forwarder)) {
    if (!ec_i.is_typed) 
      forwarded= 
	ec_i.forwarder->try_forward_consumer( this,
					      consumer_id, 
					      push_cons);
    else
      forwarded=
	ec_i.forwarder->try_forward_typed_consumer( this,
						    consumer_id,
						    ec_i.idef,
						    push_cons);
  }
  state= forwarded ? ECPS_FORWARDED : ECPS_CONNECTED;
  trace.start(100) << "Consumer " 
		   << (forwarded ? "is forwarded" : "is not forwarded")
		   << '\n';
  if (state==ECPS_CONNECTED) {
    CosEventChannelAdmin::ProxyPushSupplier::_duplicate(this);
    thr= new omni_thread( static_worker, this);
    thr->start();
  }
}

void ProxyPushSupplier_i::
disconnect_push_supplier( ENV_DEF1(env)) {
  CosEventChannelAdmin::ProxyPushSupplier_var 
    tmp(CosEventChannelAdmin::ProxyPushSupplier::_duplicate(this));
  // prevents from destructor beeing called while holding a lock
  //
  {
    omni_mutex_lock oml(guard);
    
    // BODY:
    if (state == ECPS_FORWARDED)
      ec_i.forwarder->remove_forwarded_consumer( consumer_id);
    
    if (state== ECPS_CONNECTED || state== ECPS_FORWARDED)
      ec_i.unregister_proxy_push_supplier( *this);
    if (state!= ECPS_DISCONNECTED) {
      if (state == ECPS_FORWARDED) {
	state= ECPS_DISCONNECTED;
	CosEventChannelAdmin::ProxyPushSupplier::_duplicate(this);
	(new omni_thread( finalize_worker, this))->start();
      }
      state= ECPS_DISCONNECTED;
      DEACTIVATE_OBJ( Bootstrap::boa_ptr, this);
      queue_not_empty.signal();
    }
  }
  ec_i.wakeup_checker();
}

void ProxyPushSupplier_i::
start() {
}

void ProxyPushSupplier_i::
finalize_worker( void *args) {
  ProxyPushSupplier_i &pps= *(ProxyPushSupplier_i*) args;
  
  // PRECONDITION: state==ECPS_DISCONNECTED
  //               push_consumer is not a NIL reference
  //

  try {
    pps.push_consumer->disconnect_push_consumer();
  }
  catch (...) {}
  CORBA::release(&pps);
}

void ProxyPushSupplier_i::
static_worker( void *args) {
  (*(ProxyPushSupplier_i*)args).worker();
}

void ProxyPushSupplier_i::
worker() {
  // PRECONDITION:
  // state == ECPS_CONNECTED || state == ECPS_DISCONNECTED
  //

  // pull data from event queue, and forward them to the consumer,
  // till we are disconnected, or an exception is thrown
  if (ec_i.is_typed) {
    try {
      CosTypedEventComm::TypedPushConsumer_var tpc=
	CosTypedEventComm::TypedPushConsumer::_narrow( push_consumer);
      
      if (!CORBA::is_nil(tpc)) 
	typed_consumer= tpc->get_typed_consumer();
      if (!CORBA::is_nil(typed_consumer)) {
	ec_dii= default_ec_dii_factory->create_ec_dii( typed_consumer,
						       ec_i.tpintf);
      }
    }
    catch (...) {
      typed_consumer= NULL;
    }
    if (CORBA::is_nil(typed_consumer)) {
      trace.start(100) << "Unable to obtain typed push consumer" << '\n';
    }
  }

  ECDHandler ecdh;
  for (;;) {
    {
      {
	omni_mutex_lock oml(guard);
	if ( state == ECPS_DISCONNECTED) goto out;
	if (equeue.empty()) {
	  queue_not_empty.wait();
	  if ( state == ECPS_DISCONNECTED) goto out;
	  if (equeue.empty()) continue;
	}
	ecdh= equeue.back();
	equeue.pop_back();
      }
      if (ec_i.is_typed) {
	if (ecdh->value()->_d() == ECHelper::DT_ANY) {
	  try {
	    push_consumer->push( ecdh->value()->any_value());
	  }
	  catch (...) {
	    // consumer has failed
	    goto fail;
	  }
	}
	else {
	  // forward typed data
	  if (!CORBA::is_nil(typed_consumer)) {
	    try {
	      ECHelper::TypedOperation &to= ecdh->value()->oper();
	      ec_dii->send( to.name, to.params);
	    }
	    catch (CORBA::Exception &ex) {
	      //	      trace.start(100) << ex << '\n';
	      goto fail;
	    }
	    catch (...) {
	      trace.start(100) << "Consumer has failed" << '\n';
	      // consumer has failed
	      goto fail;
	    }
	  }
	  //
	  // else: lose it, it is unlikely that we will get
	  // a typed consumer after we have previously failed to
	  //
	}
      }
      else {
	//	cerr << ecdh->value() << endl;
	if (ecdh->value()->_d() == ECHelper::DT_ANY) {
	  try {
	    push_consumer->push( ecdh->value()->any_value());
	  }
	  catch (...) {
	    // consumer has failed
	    goto fail;
	  }
	}
	else {
	  assert(0);
	  // 
	  // it is not possible, a serious program failure
	  //
	}
      }
    }    
  }

 fail:
  try {
    disconnect_push_supplier();
  }
  catch(...) {}

 out:
  trace.start(100) << "leaving ProxyPushSupplier_i::worker" << '\n';

  // we disconnected a proxy, didn't we ?
  // it would be nice to call a consumer about this situation
  try {
    push_consumer->disconnect_push_consumer();
    trace.start(100) << "disconnect_push_consumer succeeded" << '\n';
  }
  catch (CORBA::Exception &ex) {
    //    cerr << ex << endl;
    trace.start(100) << "disconnect_push_consumer failed" << '\n';
  }
  // proxy has been disconnected, drop reference count for implementation
  // possibly calling a destructor
  CORBA::release(this);
  return;
}

void ProxyPushSupplier_i::
push( ECDHandler &ecdh) {
  //  cerr << "ProxyPushSupplier_i::push" << endl;
  omni_mutex_lock oml(guard);
  if (state != ECPS_CONNECTED) return;
  if (equeue.size() >= max_queue) {
    //
    // event queue overflow
    //
    switch (overflow_action) {
    case ECAdmin::OFA_DisconnectConsumer:
      try {
	disconnect_push_supplier();
      }
      catch (...) {}
      break;
    case ECAdmin::OFA_DropOldestMessage:
      if (!equeue.empty()) // stupid, isn't it ?
	equeue.pop_back();
      break;
    case ECAdmin::OFA_DropNewMessage:
      return;
    }
  }
  equeue.push_front(ecdh);
  queue_not_empty.signal();
}

//--------------------------------------------------------ProxyPullConsumer--
// IDL interface : ProxyPullConsumer
ProxyPullConsumer_i::
ProxyPullConsumer_i( EventChannel_i &an_ec_i) : 
  ECProxy( an_ec_i),
  thr( NULL),
  state_changed( &guard),
  ec_dii(NULL),
  multi_thrs(NULL)
{
  trace.start(100) << "ProxyPullConsumer_i ctor" << '\n';
}

ProxyPullConsumer_i::
~ProxyPullConsumer_i() {
  if (multi_thrs) delete []multi_thrs;
  if (ec_dii) delete ec_dii;
  trace.start(100) << "ProxyPullConsumer_i dtor" << '\n';
}

void ProxyPullConsumer_i::
connect_pull_supplier( CosEventComm::PullSupplier_ptr pull_supp
		       ENV_DEF(env)) {
  omni_mutex_lock oml(guard);

  // PRECONDITION:
  if ( state != ECPS_NOT_CONNECTED) 
    throw CosEventChannelAdmin::AlreadyConnected();
  // it is not allowed for pull_supp to be nil reference
  ORBIX_NIL_REF_BUG( pull_supp);
  if ( CORBA::is_nil(pull_supp))
    throw CORBA::BAD_PARAM(0,CORBA::COMPLETED_NO);

  // BODY:
  pull_supplier= CosEventComm::PullSupplier::_duplicate( pull_supp);
  ec_i.register_proxy_pull_consumer( *this);

  // increment reference count for this implementation, so it should
  // last forever, at least as long as a worker thread holds its reference
  if (ec_i.is_typed) {
    ThrParams *thp= new ThrParams();
    thp->this_ptr= this;
    thp->operation_index= -1;
    CosEventChannelAdmin::ProxyPullConsumer::_duplicate(this);
    (new omni_thread( static_multi_worker, thp))->start();
  }
  else {
    CosEventChannelAdmin::ProxyPullConsumer::_duplicate(this);
    thr= new omni_thread( static_worker, this);
    thr->start();
    state= ECPS_CONNECTED;
    state_changed.signal();
  }
}

void ProxyPullConsumer_i::
start() {
}

void ProxyPullConsumer_i::
disconnect_pull_consumer( ENV_DEF1(env)) {
  {
    omni_mutex_lock oml(guard);
    
    // BODY:
    if (state==ECPS_CONNECTED) ec_i.unregister_proxy_pull_consumer( *this);
    if (state!=ECPS_DISCONNECTED) {
      state= ECPS_DISCONNECTED;
      DEACTIVATE_OBJ( Bootstrap::boa_ptr, this);
      state_changed.signal();
    }
  }
  ec_i.wakeup_checker();
}

void ProxyPullConsumer_i::
static_worker( void *args) {
  (*(ProxyPullConsumer_i*)args).worker();
}

void ProxyPullConsumer_i::
static_multi_worker( void *args) {
  ThrParams p= *((ThrParams*)args);
  //
  // free memory passed as a parameter through a heap
  //
  delete (ThrParams*) args;
  p.this_ptr->multi_worker( p.operation_index);
}

void ProxyPullConsumer_i::
multi_worker( long operation_index) {
  // this worker should only be used for typed suppliers
  //
  
  // wait for the proxy to be connected  
  {
    omni_mutex_lock oml(guard);
    while ( state == ECPS_NOT_CONNECTED)
      state_changed.wait();
    if ( state == ECPS_DISCONNECTED) goto out;
  }

  if (operation_index == -1) {
    try {
      CosTypedEventComm::TypedPullSupplier_var tps=
	CosTypedEventComm::TypedPullSupplier::_narrow( pull_supplier);
      
      if (!CORBA::is_nil(tps)) 
	typed_supplier= tps->get_typed_supplier();
      if (!CORBA::is_nil(typed_supplier)) {
	ec_dii= default_ec_dii_factory->create_ec_dii( typed_supplier,
						       ec_i.tpintf);
      }
    }
    catch (...) {
      typed_supplier= NULL;
    }
    if (CORBA::is_nil(typed_supplier)) {
      trace.start(100) << "Unable to obtain typed pull supplier" << '\n';
    }
    else {
      //
      // the first thread got a reference of typed supplier,
      // now spawn all other threads (one per typed interface operation)
      //
      for (long i=0; i<ec_i.tpintf.ops.length(); i++) {
	ThrParams *thp= new ThrParams();
	thp->this_ptr= this;
	thp->operation_index= i;
	//
	// each thread holds a reference to the implementation
	//
	CosEventChannelAdmin::ProxyPullConsumer::_duplicate(this);
	(new omni_thread( static_multi_worker, thp))->start();
      }
    }
  }

  // pull data from a pull supplier, until we are disconnected
  // or an exception is thrown
  for (;;) {
    {
      omni_mutex_lock oml(guard);
      if ( state == ECPS_DISCONNECTED) goto out;
    }
    try {
      // which operation should we call first ? ;)
      // to be fair... let's called them all at once 
      // (one operation per thread)

      if (operation_index==-1) {
	CORBA::Any *data= pull_supplier->pull();
	ec_i.push(data);
      }
      else {
	ECHelper::SeqAny sa;

	ec_dii->send_pull( ec_i.tpintf.ops[operation_index].name,
			   sa);
	
	ECHelper::ECValue *ecv= new ECHelper::ECValue();
	ECHelper::TypedOperation tp;
	tp.name= (const char *) ec_i.tpintf.ops[operation_index].name;
	tp.twoway= ec_i.tpintf.ops[operation_index].twoway;
	tp.params= sa;
	ecv->oper(tp);
	ECDHandler ecdh( new ECData(ecv), 0);
	ec_i.push(ecdh);
      }
    }
    catch (...) {
      // client failed
      goto fail;
    }
  }

 fail:
  try {
    disconnect_pull_consumer();
  }
  catch (...) {}
    
 out:
  // we disconnected a proxy, didn't we ?
  // it would be nice to call a supplier about this situation
  
  // only one thread calls "disconnect_pull_supplier"
  //

  // this call is performed in an asynchronous way, so there is no need
  // to worry about a distributed deadlock
  //
  if (operation_index==-1) {
    try {
      pull_supplier->disconnect_pull_supplier();
    }
    catch (...) {}
  }

  // proxy has been disconnected, drop reference count for implementation
  // possibly calling a destructor
  CORBA::release(this);  
}

void ProxyPullConsumer_i::
worker() {
  // this worker is designed only for untyped suppliers
  //

  // wait for the proxy to be connected  
  {
    omni_mutex_lock oml(guard);
    while ( state == ECPS_NOT_CONNECTED)
      state_changed.wait();
    if ( state == ECPS_DISCONNECTED) goto out;
  }

  // pull data from a pull supplier, until we are disconnected
  // or an exception is thrown
  for (;;) {
    {
      omni_mutex_lock oml(guard);
      if ( state == ECPS_DISCONNECTED) goto out;
    }
    CORBA::Any* data;
    try {
      data= pull_supplier->pull();
    }
    catch (...) {
      // client failed
      goto fail;
    }
    ec_i.push( data);
  }

 fail:
  try {
    disconnect_pull_consumer();
  }
  catch (...) {}
    
 out:
  // we disconnected a proxy, didn't we ?
  // it would be nice to call a supplier about this situation
  try {
    pull_supplier->disconnect_pull_supplier();
  }
  catch (...) {}

  // proxy has been disconnected, drop reference count for implementation
  // possibly calling a destructor
  CORBA::release(this);
}


//--------------------------------------------------------ProxyPullSupplier--
// IDL interface : ProxyPullSupplier
ProxyPullSupplier_i::
ProxyPullSupplier_i( EventChannel_i &an_ec_i,
		     CORBA::ULong _max_queue,
		     ECAdmin::OverflowAction _overflow_action) :
  ECProxy( an_ec_i),
  queue_not_empty(&guard),
  typedimpl(NULL),
  max_queue( _max_queue),
  overflow_action( _overflow_action),
  delete_inhibitor( ec_i.is_typed ? 2 : 1)
{
  trace.start(100) << "ProxyPullSupplier_i ctor" << '\n';

  if (ec_i.is_typed) {
    typedimpl= default_ec_dsi_factory->create_ec_dsi( ec_i.tpintf,1,*this);

    typed_corba_object= new SK_TypedProxyPullSupplier(*this);
    CONNECT_IMPL(Bootstrap::boa_ptr,corba_object);
  }
  else {
    corba_object= new SK_ProxyPullSupplier( *this);
    CONNECT_IMPL(Bootstrap::boa_ptr,corba_object);
  }
}

void ProxyPullSupplier_i::
delete_proxy( ProxyPullSupplier_i &pps) {
  int del_flag= 0;
  {
    omni_mutex_lock oml( pps.delete_guard);
    if (--pps.delete_inhibitor == 0) del_flag= 1;
  }
  if (del_flag) delete &pps;
}

ProxyPullSupplier_i::
~ProxyPullSupplier_i() {
  trace.start(100) << "~ProxyPullSupplier_i" << '\n';
}

void ProxyPullSupplier_i::
connect_pull_consumer( CosEventComm::PullConsumer_ptr pull_cons
		       ENV_DEF(env)) {
  omni_mutex_lock oml(guard);

  // PRECONDITION:
  if ( state != ECPS_NOT_CONNECTED) 
    throw CosEventChannelAdmin::AlreadyConnected();
  // it is permitted for pull_cons to be a nil reference
  ORBIX_NIL_REF_BUG( pull_cons);

  // BODY:
  pull_consumer= CosEventComm::PullConsumer::_duplicate( pull_cons);
  ec_i.register_proxy_pull_supplier( *this);
  state= ECPS_CONNECTED;
  trace.start(100) << "ProxyPullSupplier_i connected" << '\n';
}

void ProxyPullSupplier_i::
disconnect_pull_supplier( ENV_DEF1(env)) {
  {
    omni_mutex_lock oml(guard);
    
    if (state==ECPS_CONNECTED) ec_i.unregister_proxy_pull_supplier( *this);
    if (state!=ECPS_DISCONNECTED) {
      state= ECPS_DISCONNECTED;
      if (!CORBA::is_nil(pull_consumer)) {
	
	omni_mutex_lock oml2( delete_guard);
	++delete_inhibitor;
	
	(new omni_thread( finalize_worker, this))->start();
      }
      
      DEACTIVATE_OBJ( Bootstrap::boa_ptr,
		      ec_i.is_typed ? 
		      typed_corba_object : corba_object);
      
      if (ec_i.is_typed) {
	CORBA::Object_var xobj= typedimpl->get_object();
	DEACTIVATE_OBJ( Bootstrap::boa_ptr, xobj);
      }
      // wake up all pending requests
      queue_not_empty.signal();
    }
  }
  ec_i.wakeup_checker();
}

void ProxyPullSupplier_i::
finalize_worker( void *args) {
  ProxyPullSupplier_i &pps= *(ProxyPullSupplier_i*) args;
  
  // PRECONDITION: state==ECPS_DISCONNECTED
  //               pull_consumer is not a NIL reference
  //

  try {
    pps.pull_consumer->disconnect_pull_consumer();
  }
  catch (...) {}

  delete_proxy( pps);
}

CORBA::Any* ProxyPullSupplier_i::
pull( ENV_DEF1(env)) {
  for (;;) {
    if (state==ECPS_NOT_CONNECTED || state==ECPS_DISCONNECTED)
      throw CosEventComm::Disconnected();
    if (!equeue.empty()) 
      for (EQueue::reverse_iterator rit= equeue.rbegin();
	   rit != equeue.rend();
	   rit++)
	if (((**rit)->value()->_d() == ECHelper::DT_ANY)) {
	  ECDHandler v= *rit;
	  equeue.erase( (++rit).base());
	  return new CORBA::Any( (*v)->value()->any_value());
	}
    queue_not_empty.wait();
  }
}

CORBA::Any* ProxyPullSupplier_i::
try_pull( CORBA::Boolean &has_event
	  ENV_DEF(env)) {
  for (;;) {
    if (state==ECPS_NOT_CONNECTED || state==ECPS_DISCONNECTED)
      throw CosEventComm::Disconnected();
    if (!equeue.empty()) 
      for (EQueue::reverse_iterator rit= equeue.rbegin();
	   rit != equeue.rend();
	   rit++)
	if (((**rit)->value()->_d() == ECHelper::DT_ANY)) {
	  ECDHandler v= *rit;
	  equeue.erase( (++rit).base());
	  return new CORBA::Any( (*v)->value()->any_value());
	}
    // try pull
    has_event= (CORBA::Boolean) 0;
    return new CORBA::Any();
    // no event available for retrieval
  }
}

void ProxyPullSupplier_i::
push( ECDHandler &ecdh) {
  omni_mutex_lock oml(guard);
  if (state != ECPS_CONNECTED) return;
  if (equeue.size() >= max_queue) {
    //
    // event queue overflow
    //
    switch (overflow_action) {
    case ECAdmin::OFA_DisconnectConsumer:
      try {
	disconnect_pull_supplier();
      }
      catch (...) {}
      break;
    case ECAdmin::OFA_DropOldestMessage:
      if (!equeue.empty()) // stupid, isn't it ?
	equeue.pop_back();
      break;
    case ECAdmin::OFA_DropNewMessage:
      return;
    }
  }
  equeue.push_front(ecdh);
  queue_not_empty.signal();
}

CORBA::Boolean ProxyPullSupplier_i::
ec_dispatch( CORBA::Object_ptr obj, 
	     CORBA::Boolean twoway,
	     const char *operation_name,
	     ECHelper::SeqAny &params) {
  
  omni_mutex_lock oml(guard);

  // pull
  for (;;) {
    if (state!=ECPS_NOT_CONNECTED && state!=ECPS_DISCONNECTED) {
#ifdef OMNIORB
      throw CORBA::OBJECT_NOT_EXIST(0,CORBA::COMPLETED_NO);
#else
      throw CORBA::INV_OBJREF(0,CORBA::COMPLETED_NO);
#endif
    }
    if (!equeue.empty()) 
      for (EQueue::iterator rit= (equeue.end())--;
	   ;
	   rit--) {
	if (((**rit)->value()->_d() == ECHelper::DT_TYPED_OPERATION)) {
	  ECHelper::TypedOperation &to= (**rit)->value()->oper();
	  if (!strcmp(to.name,operation_name)) {
	    params= to.params;
	    equeue.erase( rit);
	    return (CORBA::Boolean) 1;
	  }
	}
	if (rit == equeue.begin()) break;
      }
    if (twoway) queue_not_empty.wait();
    else {
      // try pull
      return (CORBA::Boolean) 0;
      // no event available for retrieval
    }
  }  
  return (CORBA::Boolean) 0;
}

CORBA::Object_ptr ProxyPullSupplier_i::
get_typed_supplier() {
  return typedimpl->get_object();
}


