// $Id: ecmain.cc,v 1.4 1999/05/21 16:34:48 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/ecmain.cc,v $
//
// $Log: ecmain.cc,v $
// Revision 1.4  1999/05/21 16:34:48  mencnar
// Fixed displayed help information
//
// Revision 1.3  1999/05/04 19:23:38  mencnar
// Changed initialization of multicast layer.
//
// Revision 1.2  1999/02/24 12:38:17  mencnar
// added parameter for keep-alive calls
//
//

#include <iostream.h>
#include "ec.hh"
#include "IRread.hh"
#include "ec_dsi.hh"
#include "bootstrap.hh"

#include "trace.hh"
#include "stream_trace.hh"

#include "admin.hh"

void usage( const char *progname) {
  cerr << "Usage: " << progname << endl
       << "\t-factory_name name" << endl
       << "\t-context name" << endl
       << "\t-trace_level level" << endl
       << "\t-slave_keep_alive seconds" << endl
    ;
  exit(1);
}

int main( int argc, char **argv) {
  StreamTrace strace( cerr);
  trace.set_abstract_trace( &strace);
  trace.set_trace_level( 0);

  // Initialize an ORB and Naming Service references
  //
  Bootstrap::init(argc,argv);
  CosNaming::NamingContext_var nsroot= Bootstrap::getNS();

  //
  // Initialize multicast communcation layer
  
#ifndef NO_MCAST
  try {
    default_mcast_factory->init( argc, argv);
  }
  catch (McastFactory::InitializationError &ie) {
    cerr << "Error initializing multicast communcation layer" << endl;
    exit(1);
  }
#endif

  const char *ecfactory_name= NULL;
  const char *context_name= NULL;
  unsigned long slave_keep_alive= 60;

  for (char **pargv= argv+1; *pargv; pargv++) {
    if (!strcmp(*pargv, "-factory_name") && *(pargv+1)) {
      ecfactory_name= *(++pargv);
    }
    else if (!strcmp(*pargv, "-context") && *(pargv+1)) {
      context_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-h") || !strcmp(*pargv,"-?")) {
      usage(*argv);
    }
    else if (!strcmp(*pargv,"-trace_level") && *(pargv+1)) {
      trace.set_trace_level( atoi(*(++pargv)));
    }
    else if (!strcmp(*pargv,"-slave_keep_alive") && *(pargv+1)) {
      slave_keep_alive= atol(*(++pargv));
    }
    else {
      cerr << "Unknown option " << *pargv << endl;
      usage(*argv);
    }
  }

  if (!ecfactory_name) {
    cerr << "Undefined factory name, please use \"-factory_name\" option" << endl;
    usage(*argv);
  }
#ifndef NO_MCAST
  if (!context_name) {
    cerr << "Undefined context name, please use \"-context\" option" << endl;
    usage(*argv);
  }

  McastSourceFactory *mcast_srcf;
  try {
    mcast_srcf= default_mcast_factory->connect_as_master();
  }
  catch (McastFactory::MasterAlreadyActive &mae) {
    cerr << "Master of multicast communication group already active" << endl;
    exit(1);
  }
#endif
  trace.start(1000) << "Creating factory" << '\n';

  // 
  // Create Event Channel factory
  //
  ECAdmin::Factory_ptr ecfactory= 
    ECAdmin::Factory::_duplicate( new ECFactory_i( 
#ifndef NO_MCAST
						  mcast_srcf,
						  nsroot,
						  context_name,
						  slave_keep_alive
#endif
						  ));

  trace.start(1000) << "EventChannel Factory created" << '\n';
  CosNaming::Name nm( Bootstrap::create_name( ecfactory_name));
  try {
    nsroot->unbind(nm);
  }
  catch (CosNaming::NamingContext::NotFound &nf) {
    cerr << "unbind: not found" << endl;
  }
  catch (...) {
  }
  try {
    nsroot->bind(nm,ecfactory);
  }
  catch (CORBA::SystemException &ex) {
    cerr << "#SystemException" << endl;
    exit(1);
    return 1;
  }
  catch (CosNaming::NamingContext::AlreadyBound &) {
    cerr << "Already bound" << endl;
    return 1;
  }
  catch (CosNaming::NamingContext::CannotProceed &) {
    cerr << "Cannot proceed" << endl;
    return 1;
  }
  catch (CosNaming::NamingContext::NotFound &nf) {
    cerr << "Not found" << endl;
    return 1;
  }  
  catch (CosNaming::NamingContext::InvalidName &) {
    cerr << "Unvalid name" << endl;
  }
  catch (CORBA::UserException &) {
    cerr << "UserException" << endl;
    return 1;
  }
  catch (CORBA::Exception &) {
    cerr << "CORBA exception" << endl;
    return 1;
  }

  trace.start(5) << "Implementation is ready for service" << '\n';
  IMPL_IS_READY(Bootstrap::boa_ptr);
  return 0;
}
