// $Source: /home/iris/mencnar/ec/CVS/src/cc/factory_cl.cc,v $
//
// $Id: factory_cl.cc,v 1.2 1999/05/04 19:24:52 mencnar Exp $
//
// $Log: factory_cl.cc,v $
// Revision 1.2  1999/05/04 19:24:52  mencnar
// Added some QoS parameters: keep_alive_time and not_connected_timeout
//
// Revision 1.1  1999/02/24 13:01:56  mencnar
// *** empty log message ***
//
//

#include <iostream.h>
#include <stdlib.h>

#include <ecadmin.hh>
#include "bootstrap.hh"
#include "compatibility.hh"

void usage( char *progname) {
  cerr << progname << endl
       << "  -factory_name naming_name" << endl
       << "  -find channel_name" << endl
       << "  -find_typed channel_name" << endl
       << "  -list" << endl
       << "  -queue_size size" << endl
       << "  -overflow_action [DropOldestMessage | DropNewMessage | DisconnectConsumer]" << endl
       << "  -keep_alive_time seconds" << endl
       << "  -non_connected_timeout seconds" << endl
       << "  -create channel_name" << endl
       << "  -create_typed channel_name interface" << endl
       << "  -bind naming_name" << endl
       << "  -destroy channel_name" << endl
       << "  -destroy_typed channel_name" << endl
       << "  -shutdown" << endl
       << "  -h,-?" << endl;
  exit(1);
}

IRMOD InterfaceDef_ptr get_interface_def( const char *name) {
  IRMOD Repository_var rep;
  CosNaming::NamingContext_var rootns= Bootstrap::getNS();
  try {
    rep= IRMOD Repository::_narrow( rootns->resolve( Bootstrap::SRV_IR()));
  }
  catch (...) {
  }
  if (CORBA::is_nil(rep)) {
    cerr << "Could not find Interface Repository service" << endl;
    exit(1);
  }
  IRMOD Contained_var cntd;
  const char *intf_id= name;
  try {
    cntd= rep->lookup(name);
    if (CORBA::is_nil(cntd)) {
      cerr << "Not found" << endl;
      exit(1);
    }
  }
  catch (CORBA::SystemException &se) {
    cerr << "SystemException" << se << endl;
    exit(1);
  }
  catch (CORBA::UserException &ue) {
    cerr << "UserException" << endl;
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
    cerr << "Lookup for " << intf_id << " failed" << endl;
    exit(1);
  }

  //  cout << (const char *) orb_ptr->object_to_string(cntd) << endl;

  IRMOD IT_InterfaceDef_ptr idef;
  try {
    idef= IRMOD IT_InterfaceDef::_narrow(cntd);
    if (CORBA::is_nil(idef)) throw 1;
  }
  catch (...) {
    cerr << intf_id << " is not an interface" << endl;
    exit(1);
  }
  return idef;
}

int main( int argc, char **argv) {
  // Initialize an ORB and Naming Service references
  //
  Bootstrap::init(argc,argv);
  CosNaming::NamingContext_var nsroot= Bootstrap::getNS();

  enum { ECF_NONE, ECF_LIST, ECF_CREATE, ECF_CREATE_TYPED,
	 ECF_DESTROY, ECF_DESTROY_TYPED,
	 ECF_FIND, ECF_FIND_TYPED,
         ECF_SHUTDOWN } cmd= ECF_NONE;

  const char *ecfactory_name= NULL;
  const char *channel_name= NULL;
  const char *channel_interface= NULL;
  const char *binding_name= NULL;
  CORBA::ULong queue_size= 100;
  CORBA::ULong not_connected_timeout= 10;
  CORBA::ULong keep_alive_time= 20;
  ECAdmin::OverflowAction overflow_action= 
    ECAdmin::OFA_DropOldestMessage;

  for (char **pargv= argv+1; *pargv; pargv++) {
    if (!strcmp(*pargv, "-factory_name") && *(pargv+1)) {
      ecfactory_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-h") || !strcmp(*pargv,"-?")) {
      usage(*argv);
    }
    else if (!strcmp(*pargv,"-shutdown")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      cmd= ECF_SHUTDOWN;
    }
    else if (!strcmp(*pargv,"-list")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      cmd= ECF_LIST;
    }
    else if (!strcmp(*pargv,"-queue_size")) {
      if (!*(pargv+1)) {
	cerr << "Required paramter missing" << endl;
	usage(*argv);
      }
      queue_size= atol( *(++pargv));
    }
    else if (!strcmp(*pargv,"-not_connected_timeout")) {
      if (!*(pargv+1)) {
	cerr << "Required paramter missing" << endl;
	usage(*argv);
      }
      not_connected_timeout= atol( *(++pargv));
    }
    else if (!strcmp(*pargv,"-keep_alive_time")) {
      if (!*(pargv+1)) {
	cerr << "Required paramter missing" << endl;
	usage(*argv);
      }
      keep_alive_time= atol( *(++pargv));
    }
    else if (!strcmp(*pargv,"-overflow_action")) {
      if (!*(pargv+1)) {
	cerr << "Required paramter missing" << endl;
	usage(*argv);
      }
      ++pargv;
      if (!strcmp(*pargv,"DropOldestMessage"))
	overflow_action= ECAdmin::OFA_DropOldestMessage;
      else if (!strcmp(*pargv,"DropNewMessage"))
	overflow_action= ECAdmin::OFA_DropNewMessage;
      else if (!strcmp(*pargv,"DisconnectConsumer"))
	overflow_action= ECAdmin::OFA_DisconnectConsumer;
      else {
	cerr << "Invalid parameter for -overflow_action" << endl;
	usage(*argv);
      }
    }
    else if (!strcmp(*pargv,"-create")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      cmd= ECF_CREATE;
      channel_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-create_typed")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1) || !*(pargv+2)) {
	cerr << "Required parameters missing" << endl;
	usage(*argv);
      }
      cmd= ECF_CREATE_TYPED;
      channel_name= *(++pargv);
      channel_interface= *(++pargv);
    }
    else if (!strcmp(*pargv,"-bind")) {
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      binding_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-destroy")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      cmd= ECF_DESTROY;
      channel_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-destroy_typed")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      cmd= ECF_DESTROY_TYPED;
      channel_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-find")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      cmd= ECF_FIND;
      channel_name= *(++pargv);
    }
    else if (!strcmp(*pargv,"-find_typed")) {
      if (cmd!=ECF_NONE) {
	cerr << "Only one command allowed at a time" << endl;
	usage(*argv);
      }
      if (!*(pargv+1)) {
	cerr << "Required parameter missing" << endl;
	usage(*argv);
      }
      cmd= ECF_FIND_TYPED;
      channel_name= *(++pargv);
    }
    else {
      cerr << "Unknown option " << *pargv << endl;
      usage(*argv);
    }
  }

  if (!ecfactory_name) {
    cerr << "Undefined factory name, please use \"-factory_name\" option" << endl;
    usage(*argv);
  }
  
  ECAdmin::Factory_var factory;
  try {
    factory= ECAdmin::Factory::
      _narrow(nsroot->resolve( Bootstrap::create_name(ecfactory_name)));
  }
  catch (...) {}

  if (CORBA::is_nil(factory)) {
    cerr << "Could not find factory given by name " << ecfactory_name << endl;
    exit(1);
  }

  try {
    switch (cmd) {
    case ECF_LIST:
      {
	ECAdmin::ChannelDescriptions_var descs=
	  factory->channels();
	cout << "Found " << descs->length() << " event channel(s)" << endl;
	CORBA::ULong i;
	for (i=0; i<descs->length(); i++) {
	  cout << '#' << (unsigned long) i << endl;
	  cout << '\t' << "name:            " 
	       << (const char *) descs[i].name << endl;
	  cout << '\t' << "overflow_action: "
	       << ((descs[i].overflow_action == ECAdmin::OFA_DisconnectConsumer) ?
		   "OFA_DisconnectConsumer" :
		   ((descs[i].overflow_action == ECAdmin::OFA_DropOldestMessage) ?
		   "OFA_DropOldestMessage" :
		   ((descs[i].overflow_action == ECAdmin::OFA_DropNewMessage) ?
		   "OFA_DropNewMessage" : "?"))) << endl;
	  cout << '\t' << "max_queue_size:  "
	       << (unsigned long) descs[i].max_queue_size << endl;
	}
	ECAdmin::TypedChannelDescriptions_var tdescs=
	  factory->typed_channels();
	cerr << "Found " << tdescs->length() 
	     << " typed event channel(s)" << endl;
	for (i=0; i<tdescs->length(); i++) {
	  cout << '#' << (unsigned long) i << endl;
	  cout << '\t' << "name:            " 
	       << (const char *) tdescs[i].name << endl;
	  cout << '\t' << "overflow_action: "
	       << ((tdescs[i].overflow_action == ECAdmin::OFA_DisconnectConsumer) ?
		   "OFA_DisconnectConsumer" :
		   ((tdescs[i].overflow_action == ECAdmin::OFA_DropOldestMessage) ?
		   "OFA_DropOldestMessage" :
		   ((tdescs[i].overflow_action == ECAdmin::OFA_DropNewMessage) ?
		   "OFA_DropNewMessage" : "?"))) << endl;
	  cout << '\t' << "max_queue_size:  "
	       << (unsigned long) tdescs[i].max_queue_size << endl;
	}
      }
      break;
    case ECF_CREATE:
      {
	CosEventChannelAdmin::EventChannel_var ec=
	  factory->create_event_channel( channel_name,
					 queue_size,
					 overflow_action,
					 not_connected_timeout,
					 keep_alive_time);
	if (CORBA::is_nil(ec)) {
	  cerr << "Failed to create event channel " << channel_name << endl;
	  return 1;
	}
	else {
	  CORBA::String_var s=
	    Bootstrap::orb_ptr->object_to_string( ec);
	  cout << (const char *) s << endl;
	  if (binding_name) {
	    try {
	      CosNaming::Name name= Bootstrap::create_name(binding_name);
	      
	      try {
		nsroot->unbind( name);
	      }
	      catch (...) {}

	      nsroot->bind( name, ec);
	    }
	    catch (CORBA::Exception &ex) {
	      cerr << "Binding failed: " << ex << endl;
	      return 1;
	    }
	  }
	}
      }
      break;
    case ECF_CREATE_TYPED:
      {
	IRMOD InterfaceDef_var idef=
	  get_interface_def( channel_interface);

	CosTypedEventChannelAdmin::TypedEventChannel_var ec=
	  factory->create_typed_event_channel( idef, channel_name,
					       queue_size,
					       overflow_action,
					       not_connected_timeout,
					       keep_alive_time);
	if (CORBA::is_nil(ec)) {
	  cerr << "Failed to create typed event channel " 
	       << channel_name << endl;
	  return 1;
	}
	else {
	  CORBA::String_var s=
	    Bootstrap::orb_ptr->object_to_string( ec);
	  cout << (const char *) s << endl;
	  if (binding_name) {
	    try {
	      CosNaming::Name name= Bootstrap::create_name(binding_name);

	      try {
		nsroot->unbind( name);
	      }
	      catch (...) {}

	      nsroot->bind( name, ec);
	    }
	    catch (CORBA::Exception &ex) {
	      cerr << "Binding failed: " << ex << endl;
	      return 1;
	    }
	  }
	}
      }
      break;
    case ECF_FIND:
      {
	CosEventChannelAdmin::EventChannel_var ec=
	  factory->find_event_channel( channel_name);
	if (CORBA::is_nil(ec)) {
	  cerr << "Event channel " << channel_name 
	       << " does not exist" << endl;
	  return 1;
	}
	else {
	  CORBA::String_var s=
	    Bootstrap::orb_ptr->object_to_string( ec);
	  cout << (const char *) s << endl;
	}
      }
      break;
    case ECF_FIND_TYPED:
      {
	CosTypedEventChannelAdmin::TypedEventChannel_var ec=
	  factory->find_typed_event_channel( channel_name);
	if (CORBA::is_nil(ec)) {
	  cerr << "Typed event channel " << channel_name 
	       << " does not exist" << endl;
	  return 1;
	}
	else {
	  CORBA::String_var s=
	    Bootstrap::orb_ptr->object_to_string( ec);
	  cout << (const char *) s << endl;
	}
      }
      break;
    case ECF_DESTROY:
      {
	CosEventChannelAdmin::EventChannel_var ec=
	  factory->find_event_channel( channel_name);
	if (CORBA::is_nil(ec)) {
	  cerr << "Event channel " << channel_name 
	       << " does not exist" << endl;
	  return 1;
	}
	else {
	  ec->destroy();
	}
      }
      break;
    case ECF_DESTROY_TYPED:
      {
	CosTypedEventChannelAdmin::TypedEventChannel_var ec=
	  factory->find_typed_event_channel( channel_name);
	if (CORBA::is_nil(ec)) {
	  cerr << "Typed event channel " << channel_name 
	       << " does not exist" << endl;
	  return 1;
	}
	else {
	  ec->destroy();
	}
      }
      break;
    case ECF_SHUTDOWN:
      factory->shutdown_factory();
      break;
    }
  }
  catch (CORBA::Exception &ex) {
    cerr << ex << endl;
  }
  return 0;
}


