// $Id: slave_daemon.hh,v 1.2 1999/05/04 19:37:36 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/cc/slave_daemon.hh,v $
//
// $Log: slave_daemon.hh,v $
// Revision 1.2  1999/05/04 19:37:36  mencnar
// Added QoS parameters, keep alive calls now performed by a slave process
//
// Revision 1.1  1999/02/24 13:02:01  mencnar
// *** empty log message ***
//
//

// Event channel daemon header file
//

#ifndef __SLAVE_DAEMON_HH__
#define __SLAVE_DAEMON_HH__

#include <ecmcast.hh>
#include "utils.hh"
#include "ecdata.hh"
#include "ec_dii.hh"
#include "mcast.hh"
#include "compatibility.hh"

#include <map>
#include <set>
#include <list>

class McastSlave_i;

class SlaveForwarder {
public:
  SlaveForwarder( McastSlave_i *_mslave,
		  CosEventComm::PushConsumer_ptr push_cons,
		  ECMcast::ChannelID channel_id,
		  ECMcast::ConsumerID consumer_id,		  
		  CORBA::ULong _max_queue,
		  ECAdmin::OverflowAction _overflow_action,
		  CORBA::ULong _keep_alive_time);

  SlaveForwarder( McastSlave_i *_mslave,
		  CosEventComm::PushConsumer_ptr push_cons,
		  IRMOD InterfaceDef_ptr idef,
		  ECMcast::ChannelID channel_id,
		  ECMcast::ConsumerID consumer_id,		  
		  CORBA::ULong _max_queue,
		  ECAdmin::OverflowAction _overflow_action,
		  CORBA::ULong _keep_alive_time);

  ~SlaveForwarder();

  void push( ECDHandler &ecdh);
  void stop();
  int failed() const { return failed_flag; }
private:
  static void static_worker( void *args);
  void worker();

  CORBA::ULong keep_alive_time;
  int failed_flag;
  McastSlave_i *mslave;

  ECMcast::ChannelID channel_id;
  ECMcast::ConsumerID consumer_id;

  omni_mutex guard;
  omni_condition queue_not_empty;
  XSTD::list<ECDHandler> equeue;

  CORBA::ULong max_queue;
  ECAdmin::OverflowAction overflow_action;

  omni_thread *thr;
  CORBA::Object_var typed_consumer;
  EC_DII *ec_dii;

  CosEventComm::PushConsumer_var push_consumer;
  int active;
  int is_typed;
  ECHelper::TypedInterface tpintf; // typed event channel information
};

class ComplexID {
public:
  ECMcast::ChannelID channel_id;
  ECMcast::ConsumerID consumer_id;

  ComplexID() : channel_id(0), consumer_id(0) {}
  ComplexID( ECMcast::ChannelID _channel_id,
	     ECMcast::ConsumerID _consumer_id) :
    channel_id( _channel_id),
    consumer_id( _consumer_id) {}

  int operator == ( const ComplexID &cid) const {
    return (channel_id == cid.channel_id &&
	    consumer_id == cid.consumer_id);
  }

  int operator < ( const ComplexID &cid) const {
    if (channel_id < cid.channel_id) return 1;
    if (channel_id > cid.channel_id) return 0;
    // channel_id == cid.channel_id
    return (consumer_id < cid.consumer_id);
  }
};

class McastSlave_i : public BOA_IMPL_BASE(ECMcast,Slave) {
  typedef XSTD::map<ComplexID,SlaveForwarder*,
                    XSTD::less<ComplexID> > MapChannels;
  MapChannels channels;
  McastSinkFactory &sinkfact;
  XSTD::set<McastSink*, XSTD::less<McastSink*> > sinkset;
  ECMcast::Master_var master;

  omni_mutex guard;

  struct WorkerArgs {
    McastSlave_i *mcslave;
    McastSink *sink;

    WorkerArgs( McastSlave_i *_mcslave,
		McastSink *_sink) :
      mcslave( _mcslave),
      sink( _sink) {}
  };

  static void static_worker( void *args);
  void worker( McastSink &sink);
 public:
  McastSlave_i( McastSinkFactory &_sinkfact,
		ECMcast::Master_ptr _master);
  ~McastSlave_i();
  
  CORBA::Boolean
   forward_push_consumer( const char * mcast_channel_description,
			  ECMcast::ChannelID channel_id,
			  ECMcast::ConsumerID consumer_id,
			  CosEventChannelAdmin::ProxyPushSupplier_ptr supp,
			  CosEventComm::PushConsumer_ptr push_cons,
			  CORBA::ULong max_queue_size,
			  ECAdmin::OverflowAction oa,
			  CORBA::ULong _keep_alive_time
			  ENV_DECL(env) );
  CORBA::Boolean
    forward_typed_push_consumer( const char * mcast_channel_description,
				 ECMcast::ChannelID channel_id,
				 ECMcast::ConsumerID consumer_id,
				 IRMOD InterfaceDef_ptr idef,
				 CosEventChannelAdmin::ProxyPushSupplier_ptr supp,
				 CosEventComm::PushConsumer_ptr push_cons,
				 CORBA::ULong max_queue_size,
				 ECAdmin::OverflowAction oa,
				 CORBA::ULong _keep_alive_time
				 ENV_DECL(env) );

  void close_proxy( ECMcast::ChannelID channel_id,
		    ECMcast::ConsumerID consumer_id,
		    CORBA::Boolean call_consumer
		    ENV_DECL(env) );

  void master_restarted( ECMcast::Master_ptr new_master
			 ENV_DECL(env) );

  // additional helper methods
  ECMcast::SlaveAddresses * get_addresses();

  void slave_finished( ComplexID cid);
};

#endif // __SLAVE_DAEMON_HH__

//
