// $Id: ec.hh,v 1.3 1999/05/21 16:50:11 mencnar Exp $
//
// $Source: /home/iris/mencnar/ec/CVS/src/hh/ec.hh,v $
//
// $Log: ec.hh,v $
// Revision 1.3  1999/05/21 16:50:11  mencnar
// Added configurable "keep-alive calls"
//
// Revision 1.2  1999/01/09 14:27:22  mencda
// *** empty log message ***
//
// Revision 1.1.1.1  1998/12/13 23:27:16  mencda
// Imported sources
//
//

#ifndef __EC_HH__
#define __EC_HH__

#include <iostream.h>
#include "ecdata.hh"
#include <list>
#include <set>
#include <map>
#include "ifr.hh"
#include "compatibility.hh"
#include "ecadmin.hh"
#include "ec_dsi.hh"
#include "ec_dii.hh"
#include "forward.hh"
#include "unique_ids.hh"


// forward declarations of tie-like implementation classes
//
class ConsumerAdmin_i;
class SupplierAdmin_i;
class ProxyPushConsumer_i;
class ProxyPushSupplier_i;
class ProxyPullConsumer_i;
class ProxyPullSupplier_i;

class EventChannel_i;
class ECProxy;
class ECRef;

class MyECController {
public:
  virtual void channel_disposed( EventChannel_i *)= 0;
  // an event channel has beeen deactivated
  // the controller object should not perform any actions 
  // (deleting implmentation, etc. )
  // except for destroying all corba references associated with this object
  // (if it has such), than it should assume object nonexistence,
  // the actual destruction of the object may be delayed, but it will
  // be performed by EventChannel_i implementation and associated objects,
  // therefore instances of EventChannel_i object should not be created
  // otherwise than on the heap
  // 
};

class EventChannel_i {
public:
  const int is_typed;
  const CORBA::ULong not_connected_timeout;
  const CORBA::ULong keep_alive_time;
private:
  int is_active;
  omni_mutex checker_guard;
  omni_condition checker_cond;
  typedef XSTD::list<ECProxy*> CheckerProxies;
  CheckerProxies checker_proxies;
  ECRef* checker_ec_instance;
  static void static_checker_worker( void *args);
  void checker_worker();

  static UniqueID<CORBA::ULong> channel_id_dispenser;
  MyECController *eccontr;

  ConsumerAdmin_i* consumer_admin;
  SupplierAdmin_i* supplier_admin;

  GeneralLock proxy_guard;

  // 
  typedef XSTD::set<ProxyPushConsumer_i*,
                    XSTD::less<ProxyPushConsumer_i*> > SetPPushCons;  
  typedef XSTD::set<ProxyPushSupplier_i*,
                    XSTD::less<ProxyPushSupplier_i*> > SetPPushSupp;
  typedef XSTD::set<ProxyPullConsumer_i*,
                    XSTD::less<ProxyPullConsumer_i*> > SetPPullCons;
  typedef XSTD::set<ProxyPullSupplier_i*,
                    XSTD::less<ProxyPullSupplier_i*> > SetPPullSupp;
  typedef XSTD::map<CORBA::ULong,ProxyPushSupplier_i*,
                    XSTD::less<CORBA::ULong> > MapIDProxy;

  SetPPushCons proxy_push_consumers;   
  SetPPushSupp proxy_push_suppliers;
  SetPPullCons proxy_pull_consumers;
  SetPPullSupp proxy_pull_suppliers;
  MapIDProxy id_to_proxy_push_supplier;
  CORBA::ULong channel_id;

  CORBA::ULong default_max_queue;
  ECAdmin::OverflowAction default_overflow_action;

  ~EventChannel_i();
  // a private destructor prevents creating automatic instances
  // of this class
  //
  friend class SK_EventChannel;
  friend class SK_TypedEventChannel;

public:
  EventChannel_i( CORBA::ULong _max_queue,
		  ECAdmin::OverflowAction _overflow_action,
		  CORBA::ULong _not_connected_timeout,
		  CORBA::ULong _keep_alive_time,
		  AbstractForwarderInstance _forwarder,
		  MyECController *_eccontr);
  // untyped event channel constructor

  EventChannel_i( CORBA::ULong _max_queue,
		  ECAdmin::OverflowAction _overflow_action,
		  CORBA::ULong _not_connected_timeout,
		  CORBA::ULong _keep_alive_time,
		  AbstractForwarderInstance _forwarder,
		  IRMOD InterfaceDef_ptr idef,
		  MyECController *_eccontr);
  // typed event channel constructor

  ECHelper::TypedInterface tpintf; // typed event channel information

  UniqueID<CORBA::ULong> consumer_id_dispenser;
  CosEventChannelAdmin::EventChannel_ptr corba_object;
  CosTypedEventChannelAdmin::TypedEventChannel_ptr typed_corba_object;
  IRMOD InterfaceDef_var idef;
  AbstractForwarderInstance forwarder;

  CORBA::ULong get_max_queue() {
    return default_max_queue;
  }

  ECAdmin::OverflowAction get_overflow_action() {
    return default_overflow_action;
  }

  void check_interface_name( const char *name);

  void insert_checker_proxy( ECProxy*);
  void wakeup_checker();

  // IDL interface : EventChannel
  CosEventChannelAdmin::ConsumerAdmin_ptr for_consumers();
  CosEventChannelAdmin::SupplierAdmin_ptr for_suppliers();
  void destroy();

  // IDL interface : TypedEventChannel
  CosTypedEventChannelAdmin::TypedConsumerAdmin_ptr for_typed_consumers();
  CosTypedEventChannelAdmin::TypedSupplierAdmin_ptr for_typed_suppliers();

  // IDL interface : ConsumerAdmin
  CosEventChannelAdmin::ProxyPushSupplier_ptr obtain_push_supplier();
  CosEventChannelAdmin::ProxyPullSupplier_ptr obtain_pull_supplier();

  // IDL interface : TypedConsumerAdmin
  CosEventChannelAdmin::ProxyPushSupplier_ptr 
    obtain_typed_push_supplier( const char *intf);
  
  CosTypedEventChannelAdmin::TypedProxyPullSupplier_ptr 
    obtain_typed_pull_supplier( const char *intf);

  // IDL interface : SupplierAdmin
  CosEventChannelAdmin::ProxyPushConsumer_ptr obtain_push_consumer();
  CosEventChannelAdmin::ProxyPullConsumer_ptr obtain_pull_consumer();

  // IDL interface : TypedSupplierAdmin
  CosTypedEventChannelAdmin::TypedProxyPushConsumer_ptr
    obtain_typed_push_consumer( const char *intf);
  CosEventChannelAdmin::ProxyPullConsumer_ptr
    obtain_typed_pull_consumer( const char *intf);

  // Implementation
  EventChannel_i( const char *a_name);  

  void register_proxy_push_consumer( ProxyPushConsumer_i &);
  void unregister_proxy_push_consumer( ProxyPushConsumer_i &);
  void register_proxy_push_supplier( ProxyPushSupplier_i &);
  void unregister_proxy_push_supplier( ProxyPushSupplier_i &);

  void register_proxy_pull_consumer( ProxyPullConsumer_i &);
  void unregister_proxy_pull_consumer( ProxyPullConsumer_i &);
  void register_proxy_pull_supplier( ProxyPullSupplier_i &);
  void unregister_proxy_pull_supplier( ProxyPullSupplier_i &);

  void push( ECDHandler &ecdh);
  void push( const CORBA::Any &data);
  void push( CORBA::Any *data);

  CosEventChannelAdmin::ProxyPushSupplier_ptr
    find_proxy_push_supplier( CORBA::ULong consumer_id);
 private:

  DISABLE_COPY_AND_ASSIGN(EventChannel_i)
};

// this class holds references of an event channel
// the side effect: duplicates reference count of corba object in ctor
//                  and releases it in dtor
class ECRef {
public:
  ECRef() :
    ec_i(NULL) {}

  ECRef( EventChannel_i &_ec_i) : 
    ec_i( &_ec_i) {

    if (!ec_i->is_typed) 
      corba_object= CosEventChannelAdmin::EventChannel::
      _duplicate( ec_i->corba_object);
    else 
      typed_corba_object= CosTypedEventChannelAdmin::TypedEventChannel::
      _duplicate( ec_i->typed_corba_object);
  }
   
  EventChannel_i *ec_i;
  CosEventChannelAdmin::EventChannel_var corba_object;
  CosTypedEventChannelAdmin::TypedEventChannel_var typed_corba_object;
};

//--------------------------------------------------------------------------
class SK_EventChannel
  : public BOA_IMPL_BASE(CosEventChannelAdmin,EventChannel) 
{
  EventChannel_i &i;
 public:
  SK_EventChannel( EventChannel_i &_i) : i(_i) {}

  // IDL interface : EventChannel
  CosEventChannelAdmin::ConsumerAdmin_ptr for_consumers( ENV_DECL1(env)) {
    return i.for_consumers();
  }
  CosEventChannelAdmin::SupplierAdmin_ptr for_suppliers( ENV_DECL1(env)) {
    return i.for_suppliers();
  }
  void destroy( ENV_DECL1(env)) {
    i.destroy();
  }

  ~SK_EventChannel() {
    delete &i;
  }
};

class SK_TypedEventChannel
  : public BOA_IMPL_BASE(CosTypedEventChannelAdmin,TypedEventChannel)
{
  EventChannel_i &i;
 public:
  SK_TypedEventChannel( EventChannel_i &_i) : i(_i) {}

  // IDL interface : TypedEventChannel
  CosTypedEventChannelAdmin::TypedConsumerAdmin_ptr for_consumers( ENV_DECL1(env)) {
    return i.for_typed_consumers();
  }
  CosTypedEventChannelAdmin::TypedSupplierAdmin_ptr for_suppliers( ENV_DECL1(env)) {
    return i.for_typed_suppliers();
  }
  void destroy( ENV_DECL1(env)) {
    i.destroy();
  }

  ~SK_TypedEventChannel() {
    delete &i;
  }
};


//---------------------------------------------------------------------------

class ConsumerAdmin_i
{
 public:
  const int is_typed;
  CosEventChannelAdmin::ConsumerAdmin_ptr corba_object;
  CosTypedEventChannelAdmin::TypedConsumerAdmin_ptr typed_corba_object;

  // IDL interface
  CosEventChannelAdmin::ProxyPushSupplier_ptr 
    obtain_push_supplier() {

    return ec_i.obtain_push_supplier();
  }
  CosEventChannelAdmin::ProxyPullSupplier_ptr 
    obtain_pull_supplier() {

    return ec_i.obtain_pull_supplier();
  }

  CosTypedEventChannelAdmin::TypedProxyPullSupplier_ptr 
    obtain_typed_pull_supplier(const char *intf) {

    return ec_i.obtain_typed_pull_supplier( intf);
  }

  CosEventChannelAdmin::ProxyPushSupplier_ptr
    obtain_typed_push_supplier( const char *intf) {

    return ec_i.obtain_typed_push_supplier( intf);
  }  

  // Implementation
  ConsumerAdmin_i( EventChannel_i &an_ec_i);
  ~ConsumerAdmin_i() {
    cerr << "~ConsumerAdmin_i" << endl;
  }
 private:
  EventChannel_i &ec_i;

  DISABLE_COPY_AND_ASSIGN(ConsumerAdmin_i)
};

class SK_ConsumerAdmin
  : public BOA_IMPL_BASE(CosEventChannelAdmin,ConsumerAdmin)
{
  ConsumerAdmin_i &i;
 public:
  SK_ConsumerAdmin( ConsumerAdmin_i &_i) : i(_i) {}
  // IDL interface
  CosEventChannelAdmin::ProxyPushSupplier_ptr 
    obtain_push_supplier( ENV_DECL1(env)) {

    return i.obtain_push_supplier();
  }
  CosEventChannelAdmin::ProxyPullSupplier_ptr 
    obtain_pull_supplier( ENV_DECL1(env)) {

    return i.obtain_pull_supplier();
  }
  ~SK_ConsumerAdmin() {
    delete &i;
  }
};

class SK_TypedConsumerAdmin
  : public BOA_IMPL_BASE(CosTypedEventChannelAdmin,TypedConsumerAdmin)
{
  ConsumerAdmin_i &i;
 public:
  SK_TypedConsumerAdmin( ConsumerAdmin_i &_i) : i(_i) {}
  // IDL interface
  CosEventChannelAdmin::ProxyPushSupplier_ptr 
    obtain_push_supplier( ENV_DECL1(env)) {

    return i.obtain_push_supplier();
  }
  CosEventChannelAdmin::ProxyPullSupplier_ptr 
    obtain_pull_supplier( ENV_DECL1(env)) {

    return i.obtain_pull_supplier();
  }
  CosTypedEventChannelAdmin::TypedProxyPullSupplier_ptr 
    obtain_typed_pull_supplier(const char *intf
			       ENV_DECL(env)) {

    return i.obtain_typed_pull_supplier( intf);
  }

  CosEventChannelAdmin::ProxyPushSupplier_ptr
    obtain_typed_push_supplier( const char *intf
				ENV_DECL(env)) {

    return i.obtain_typed_push_supplier( intf);
  }  
  ~SK_TypedConsumerAdmin() {
    delete &i;
  }
};

//---------------------------------------------------------------------------

class SupplierAdmin_i
{
 public:
  const int is_typed;
  CosEventChannelAdmin::SupplierAdmin_ptr corba_object;
  CosTypedEventChannelAdmin::TypedSupplierAdmin_ptr typed_corba_object;

  // IDL interface
  CosEventChannelAdmin::ProxyPushConsumer_ptr 
    obtain_push_consumer() {

    return ec_i.obtain_push_consumer();
  }
  CosEventChannelAdmin::ProxyPullConsumer_ptr 
    obtain_pull_consumer() {

    return ec_i.obtain_pull_consumer();
  }

  CosTypedEventChannelAdmin::TypedProxyPushConsumer_ptr 
    obtain_typed_push_consumer( const char *key) {

    return ec_i.obtain_typed_push_consumer( key);
  }

  CosEventChannelAdmin::ProxyPullConsumer_ptr
    obtain_typed_pull_consumer( const char *key) {

    return ec_i.obtain_typed_pull_consumer( key);
  }

  // Implementation
  SupplierAdmin_i( EventChannel_i &an_ec_i);
  ~SupplierAdmin_i() {
    cerr << "~SupplierAdmin_i" << endl;
  }
 private:
  EventChannel_i &ec_i;

  DISABLE_COPY_AND_ASSIGN(SupplierAdmin_i)
};

class SK_SupplierAdmin
  : public BOA_IMPL_BASE(CosEventChannelAdmin,SupplierAdmin)
{
  SupplierAdmin_i &i;
 public:
  SK_SupplierAdmin( SupplierAdmin_i &_i) : i(_i) {}

  CosEventChannelAdmin::ProxyPushConsumer_ptr 
    obtain_push_consumer( ENV_DECL1(env)) {

    return i.obtain_push_consumer();
  }
  CosEventChannelAdmin::ProxyPullConsumer_ptr 
    obtain_pull_consumer( ENV_DECL1(env)) {

    return i.obtain_pull_consumer();
  }  
  ~SK_SupplierAdmin() {
    delete &i;
  }
};

class SK_TypedSupplierAdmin
  : public BOA_IMPL_BASE(CosTypedEventChannelAdmin,TypedSupplierAdmin)
{
  SupplierAdmin_i &i;
 public:
  SK_TypedSupplierAdmin( SupplierAdmin_i &_i) : i(_i) {}

  CosEventChannelAdmin::ProxyPushConsumer_ptr 
    obtain_push_consumer( ENV_DECL1(env)) {

    return i.obtain_push_consumer();
  }
  CosEventChannelAdmin::ProxyPullConsumer_ptr 
    obtain_pull_consumer( ENV_DECL1(env)) {

    return i.obtain_pull_consumer();
  }

  CosTypedEventChannelAdmin::TypedProxyPushConsumer_ptr
    obtain_typed_push_consumer( const char *intf
				ENV_DECL(env)) {
    return i.obtain_typed_push_consumer( intf);
  }

  CosEventChannelAdmin::ProxyPullConsumer_ptr
    obtain_typed_pull_consumer( const char *intf
				ENV_DECL(env)) {
    return i.obtain_typed_pull_consumer( intf);
  }
  ~SK_TypedSupplierAdmin() {
    delete &i;
  }
};

enum ECProxyState { 
  ECPS_NOT_CONNECTED, ECPS_CONNECTED, ECPS_DISCONNECTED,
  ECPS_FORWARDED
};

//-------------------------------------------------------------------ECProxy--
//
class ECProxy {
protected:
  ECProxyState state;
  omni_mutex guard;
  ECRef ecref;
  EventChannel_i &ec_i;
  TimeStamp timestamp;
  
  ECProxy( EventChannel_i &an_ec_i) : 
    state( ECPS_NOT_CONNECTED),
    ecref( an_ec_i),
    ec_i( an_ec_i),
    timestamp( default_tsf->instantiate()) {}
public:
  ECProxyState get_state() {
    omni_mutex_lock oml(guard);
    return state;
  }

  TimeStamp get_timestamp() {
    omni_mutex_lock oml(guard);
    return timestamp;
  }

  void touch() {
    omni_mutex_lock oml(guard);
    my_touch();
  }

  virtual void duplicate_proxy()= 0;
  virtual void release_proxy()= 0;
  virtual void deactivate_proxy()= 0;

  virtual CORBA::Object_ptr get_client_reference()= 0;
  // returns a duplicated reference of a connected client
  // 

protected:
  void my_touch() {
    timestamp= default_tsf->instantiate();
  }
};

//-------------------------------------------------------ProxyPushConsumer_i--
//
class ProxyPushConsumer_i : public EC_Dispatcher,
			    public ECProxy
{
 public:
  CosEventChannelAdmin::ProxyPushConsumer_ptr corba_object;
  CosTypedEventChannelAdmin::TypedProxyPushConsumer_ptr typed_corba_object;

  // IDL interface
  void connect_push_supplier( CosEventComm::PushSupplier_ptr push_supp);
  void push( const CORBA::Any &data);
  void disconnect_push_consumer();
  CORBA::Object_ptr get_typed_consumer();

  // Implementation
  ProxyPushConsumer_i( EventChannel_i &an_ec_i);
  ~ProxyPushConsumer_i() {
    cerr << "ProxyPushConsumer_i dtor" << endl;
  }
  void start() {}

  CORBA::Boolean ec_dispatch( CORBA::Object_ptr obj, 
			      CORBA::Boolean twoway,
			      const char *operation_name,
			      ECHelper::SeqAny &params);
  void ec_dsi_disposed() {
    delete_proxy(*this);
  }
  
  static void delete_proxy( ProxyPushConsumer_i &ppc);

  virtual void duplicate_proxy() {
    omni_mutex_lock oml(delete_guard);
    delete_inhibitor++;
  }

  virtual void release_proxy() {
    delete_proxy( *this);
  }

  virtual void deactivate_proxy() {
    disconnect_push_consumer();
  }

  virtual CORBA::Object_ptr get_client_reference() {
    return CosEventComm::PushSupplier::_duplicate(push_supplier);
  }

 private:
  CosEventComm::PushSupplier_var push_supplier;
  EC_DSI *typedimpl;  
  omni_mutex delete_guard;
  int delete_inhibitor;

  static void finalize_worker( void *args);

  DISABLE_COPY_AND_ASSIGN(ProxyPushConsumer_i)
};

class SK_ProxyPushConsumer
  : public BOA_IMPL_BASE(CosEventChannelAdmin,ProxyPushConsumer)
{
  ProxyPushConsumer_i &i;
 public:
  SK_ProxyPushConsumer( ProxyPushConsumer_i &_i) : i(_i) {}

  void connect_push_supplier( CosEventComm::PushSupplier_ptr push_supp
			      ENV_DECL(env)) {
    i.connect_push_supplier( push_supp);
  }
  void push( const CORBA::Any &data
	     ENV_DECL(env)) {
    i.push( data);
  }

  void disconnect_push_consumer( ENV_DECL1(env)) {
    i.disconnect_push_consumer();
  }

  ~SK_ProxyPushConsumer() {
    ProxyPushConsumer_i::delete_proxy(i);
  }
};

class SK_TypedProxyPushConsumer
  : public BOA_IMPL_BASE(CosTypedEventChannelAdmin,TypedProxyPushConsumer)
{
  ProxyPushConsumer_i &i;
 public:
  SK_TypedProxyPushConsumer( ProxyPushConsumer_i &_i) : i(_i) {}

  void connect_push_supplier( CosEventComm::PushSupplier_ptr push_supp
			      ENV_DECL(env)) {
    i.connect_push_supplier( push_supp);
  }
  void push( const CORBA::Any &data
	     ENV_DECL(env)) {
    i.push( data);
  }

  void disconnect_push_consumer( ENV_DECL1(env)) {
    i.disconnect_push_consumer();
  }

  CORBA::Object_ptr get_typed_consumer( ENV_DECL1(env)) {
    return i.get_typed_consumer();
  }

  ~SK_TypedProxyPushConsumer() {
    ProxyPushConsumer_i::delete_proxy(i);
  }
};

//-------------------------------------------------------ProxyPushSupplier_i--
//

class ProxyPushSupplier_i
  : public BOA_IMPL_BASE(CosEventChannelAdmin,ProxyPushSupplier),
    public ECProxy

{
 public:
  // IDL interface
  void connect_push_consumer( CosEventComm::PushConsumer_ptr push_cons
			      ENV_DECL(env));
  void disconnect_push_supplier( ENV_DECL1(env));
  
  // Implementation
  ProxyPushSupplier_i( EventChannel_i &an_ec_i,
		       CORBA::ULong _max_queue,
		       ECAdmin::OverflowAction _overflow_action);

  ~ProxyPushSupplier_i() {
    if (ec_dii) delete ec_dii;
    cerr << "ProxyPushSupplier_i dtor" << endl;
  }
  void start();
  void push( ECDHandler &ecdh);
  CORBA::ULong get_id() const { return consumer_id; }

  virtual void duplicate_proxy() {
    CosEventChannelAdmin::ProxyPushSupplier::_duplicate( this);
  }

  virtual void release_proxy() {
    CORBA::release( this);
  }

  virtual void deactivate_proxy() {
    disconnect_push_supplier();
  }

  virtual CORBA::Object_ptr get_client_reference() {
    return CosEventComm::PushConsumer::_duplicate(push_consumer);
  }

 private:
  static void static_worker( void *args);
  void worker();

  omni_condition queue_not_empty;
  typedef XSTD::list<ECDHandler> EQueue;
  EQueue equeue;

  CORBA::ULong max_queue;
  ECAdmin::OverflowAction overflow_action;

  omni_thread *thr;
  CORBA::Object_var typed_consumer;
  EC_DII *ec_dii;

  CosEventComm::PushConsumer_var push_consumer;
  int forwarded;
  CORBA::ULong consumer_id;

  static void finalize_worker( void *args);

  DISABLE_COPY_AND_ASSIGN(ProxyPushSupplier_i)
};

class ProxyPullConsumer_i
  : public BOA_IMPL_BASE(CosEventChannelAdmin,ProxyPullConsumer),
    public ECProxy
{
  struct ThrParams {
    ProxyPullConsumer_i *this_ptr;
    long operation_index;
  };

 public:
  // IDL interface
  void connect_pull_supplier( CosEventComm::PullSupplier_ptr pull_supp
			      ENV_DECL(env));
  void disconnect_pull_consumer( ENV_DECL1(env));

  // Implementation
  ProxyPullConsumer_i( EventChannel_i &an_ec_i);
  ~ProxyPullConsumer_i();
  void start();

  virtual void duplicate_proxy() {
    CosEventChannelAdmin::ProxyPullConsumer::_duplicate( this);
  }

  virtual void release_proxy() {
    CORBA::release( this);
  }

  virtual void deactivate_proxy() {
    disconnect_pull_consumer();
  }

  virtual CORBA::Object_ptr get_client_reference() {
    return CosEventComm::PullSupplier::_duplicate(pull_supplier);
  }

 private:
  static void static_worker( void *args);
  static void static_multi_worker( void *args);
  void worker();
  void multi_worker( long operation_index);

  omni_condition state_changed;
  omni_thread *thr;
  omni_thread **multi_thrs;
  CosEventComm::PullSupplier_var pull_supplier;

  CORBA::Object_var typed_supplier;
  EC_DII *ec_dii;

  DISABLE_COPY_AND_ASSIGN(ProxyPullConsumer_i)
};

//-------------------------------------------------------ProxyPullSupplier_i--
//
class ProxyPullSupplier_i : public EC_Dispatcher,
			    public ECProxy
{
 public:
  CosEventChannelAdmin::ProxyPullSupplier_ptr corba_object;
  CosTypedEventChannelAdmin::TypedProxyPullSupplier_ptr typed_corba_object;

  // IDL interface
  void connect_pull_consumer( CosEventComm::PullConsumer_ptr pull_cons
			      ENV_DECL(env));

  CORBA::Any* pull( ENV_DECL1(env));
  CORBA::Any* try_pull( CORBA::Boolean &has_event
			ENV_DECL(env));
  void disconnect_pull_supplier( ENV_DECL1(env));

  CORBA::Object_ptr get_typed_supplier();
 
  // Implementation
  ProxyPullSupplier_i( EventChannel_i &an_ec_i,
		       CORBA::ULong _max_queue,
		       ECAdmin::OverflowAction _overflow_action );
  ~ProxyPullSupplier_i();
  void start() {}

  void push( ECDHandler &ecdh);

  CORBA::Boolean ec_dispatch( CORBA::Object_ptr obj, 
			      CORBA::Boolean twoway,
			      const char *operation_name,
			      ECHelper::SeqAny &params);

  void ec_dsi_disposed() {
    delete_proxy(*this);
  }
  
  static void delete_proxy( ProxyPullSupplier_i &ppc);

  virtual void duplicate_proxy() {
    omni_mutex_lock oml(delete_guard);
    delete_inhibitor++;
  }

  virtual void release_proxy() {
    delete_proxy( *this);
  }

  virtual void deactivate_proxy() {
    disconnect_pull_supplier();
  }

  virtual CORBA::Object_ptr get_client_reference() {
    return CosEventComm::PullConsumer::_duplicate(pull_consumer);
  }

 private:
  CosEventComm::PullConsumer_var pull_consumer;
  EC_DSI *typedimpl;  

  omni_condition queue_not_empty;
  typedef XSTD::list<ECDHandler> EQueue;
  EQueue equeue;

  CORBA::ULong max_queue;
  ECAdmin::OverflowAction overflow_action;

  omni_mutex delete_guard;
  int delete_inhibitor;

  static void finalize_worker( void *args);

  DISABLE_COPY_AND_ASSIGN(ProxyPullSupplier_i)
};

class SK_ProxyPullSupplier 
  : public BOA_IMPL_BASE(CosEventChannelAdmin,ProxyPullSupplier)
{
  ProxyPullSupplier_i &i;
 public:
  SK_ProxyPullSupplier( ProxyPullSupplier_i &_i) : i(_i) {}

  void connect_pull_consumer( CosEventComm::PullConsumer_ptr pull_cons
			      ENV_DECL(env)) {
    i.connect_pull_consumer( pull_cons);
  }

  CORBA::Any* pull( ENV_DECL1(env)) {
    return i.pull();
  }
  CORBA::Any* try_pull( CORBA::Boolean &has_event
			ENV_DECL(env)) {
    return i.try_pull( has_event);
  }
  void disconnect_pull_supplier( ENV_DECL1(env)) {
    i.disconnect_pull_supplier();
  }

  ~SK_ProxyPullSupplier() {
    ProxyPullSupplier_i::delete_proxy(i);
  }
};

class SK_TypedProxyPullSupplier 
  : public BOA_IMPL_BASE(CosTypedEventChannelAdmin,TypedProxyPullSupplier)
{
  ProxyPullSupplier_i &i;
 public:
  SK_TypedProxyPullSupplier( ProxyPullSupplier_i &_i) : i(_i) {}

  void connect_pull_consumer( CosEventComm::PullConsumer_ptr pull_cons
			      ENV_DECL(env)) {
    i.connect_pull_consumer( pull_cons);
  }

  CORBA::Any* pull( ENV_DECL1(env)) {
    return i.pull();
  }
  CORBA::Any* try_pull( CORBA::Boolean &has_event
			ENV_DECL(env)) {
    return i.try_pull( has_event);
  }
  void disconnect_pull_supplier( ENV_DECL1(env)) {
    i.disconnect_pull_supplier();
  }

  CORBA::Object_ptr get_typed_supplier( ENV_DECL1(env)) {
    return i.get_typed_supplier();
  }

  ~SK_TypedProxyPullSupplier() {
    ProxyPullSupplier_i::delete_proxy(i);
  }
};

#endif // __EC_HH__

//
