#ifndef __MCAST_HH__
#define __MCAST_HH__

// $Id: mcast.hh,v 1.1 1999/01/09 14:27:23 mencda Exp $
//
// $Source: /home/mencda/CVSROOT/ec/src/hh/mcast.hh,v $
//
// $Log: mcast.hh,v $
// Revision 1.1  1999/01/09 14:27:23  mencda
// *** empty log message ***
//
//

//
// A set of abstract interfaces to glue together an EventChannel
// and a mutlicast communication layer.
// Each communication layer must support these interfaces.
//

#include "locks.hh"

class McastSourceFactory;
class McastSinkFactory;
class McastSource;
class McastSink;

class McastFactory {
public:
  class InitializationError {
  public:
    // why ?
  };

  virtual void init( int & argc, char ** &argv)= 0;
  
  class MasterAlreadyActive {};
  class NoMasterActive {};

  virtual McastSourceFactory *connect_as_master()= 0;
  virtual McastSinkFactory *connect_as_slave()= 0;
  virtual ~McastFactory() {}
};

extern McastFactory *default_mcast_factory;

class McastSlaveConnectedCallback {
public:
  virtual void slave_connected()= 0;
};

class McastSourceFactory {
public:
  virtual McastSource *create( McastSlaveConnectedCallback *cback,
			       int new_channel_hint)= 0;
  virtual ~McastSourceFactory() {}
};

class McastSource {
public:
  virtual char *get_unique_description() { return new char[0]; }
  // returns a unique description of the communicatio channel
  // that can be used to initialize sinks
  // the memory allocated for the return value should be reclaimed
  // using delete [] operator
  //

  GeneralLock *guard;
  // before call to start_message a write lock for "guard" object should
  // be acquired, then it should be release after call to "end_message",
  // this is to avoid a clash among multiple threads using the same
  // instance of McastSource object
  //

  McastSource( GeneralLock *_guard) :
    guard( _guard) {}

  virtual void start_message()= 0;
  virtual void end_message()= 0;

  virtual void write( void *data, unsigned long len)= 0;
  virtual ~McastSource() {}
};


class McastSinkFactory {
public:
  virtual McastSink *create( const char *unique_description)= 0;
  virtual ~McastSinkFactory() {}
};

class McastSink {
public:
  class LostData {}; // exception thrown by read method
  
  virtual void read( void *buf, unsigned long buflen)= 0;
  virtual ~McastSink() {}
};

#endif // __MCAST_HH__

// end of $File$
