#ifndef __TRACE_HH__
#define __TRACE_HH__
// $Id$
//
// $Source$
//
// $Log$
//

class AbstractTrace {
public:
  virtual void out_string( const char *)= 0;
  virtual void out_char( char)= 0;
  virtual void out_int( int)= 0;
  virtual void out_unsigned( unsigned)= 0;
  virtual void out_short( short)= 0;
  virtual void out_unsigned_short( unsigned short)= 0;
  virtual void out_long( long)= 0;
  virtual void out_unsigned_long( unsigned long)= 0;
};

class TraceSink {
  AbstractTrace *atrace;
  unsigned trace_level;
  unsigned current_level;
  int tracing;
public:
  TraceSink() : atrace(0), trace_level(0), current_level(0), tracing(0) {}

  void set_abstract_trace( AbstractTrace *at) {
    atrace= at;
  }
   
  void set_trace_level( unsigned new_level) {
    trace_level= new_level;
  }

  TraceSink& start( unsigned level) {
    tracing= ((current_level= level) < trace_level);
    return *this;
  }
  TraceSink& operator <<( const char *s) {
    if (atrace && tracing) atrace->out_string( s);
    return *this;
  }
  TraceSink& operator <<( char c) {
    if (atrace && tracing) atrace->out_char( c);
    return *this;
  }
  TraceSink& operator <<( int i) {
    if (atrace && tracing) atrace->out_int( i);
    return *this;
  }
  TraceSink& operator <<( unsigned u) {
    if (atrace && tracing) atrace->out_unsigned( u);
    return *this;
  }
  TraceSink& operator <<( short s) {
    if (atrace && tracing) atrace->out_short( s);
    return *this;
  }
  TraceSink& operator <<( unsigned short us) {
    if (atrace && tracing) atrace->out_unsigned_short( us);
    return *this;
  }
  TraceSink& operator <<( long l) {
    if (atrace && tracing) atrace->out_long( l);
    return *this;
  }
  TraceSink& operator <<( unsigned long ul) {
    if (atrace && tracing) atrace->out_unsigned_long( ul);
    return *this;
  }
};

extern TraceSink trace;

#endif // __TRACE_HH__

// end of $File$
