// $Id: mtp.cc,v 1.2 1998/11/03 18:06:07 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtp.cc,v $
//
// $Log: mtp.cc,v $
// Revision 1.2  1998/11/03 18:06:07  mencda
// token request/confirm routines
//
//

#ifndef NO_MCAST

#include <mtime.hh>

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>

#ifdef __WIN32__
//#define INCL_WINSOCK_API_PROTOTYPES 1
#include <winsock2.h>
#include <windows.h>
typedef int socklen_t;
#else
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#endif

#include <string.h>
//#include <getopt.h>

#ifdef __WIN32__
#include <strstrea.h>
#else
#include <strstream.h>
#endif
#include "mtp.hh"

#ifdef __sunos__
extern "C" struct hostent *gethostbyname(const char *name);
extern "C" int gethostname(char *name, int namelen);
#endif

MTPConnectionIdMgr mtpidmgr;
PacketMgrFactory packetmgrfactory;

int resolve_host( const char *hostname,
		  sockaddr &sa,
		  unsigned &sa_len,
		  unsigned port) {

  sockaddr_in &myaddr= (sockaddr_in&) sa;

  hostent *h= gethostbyname( hostname);
  if (h== NULL) return -1;
  myaddr.sin_family = AF_INET;
  myaddr.sin_port = htons(port);

  memcpy((void *)&myaddr.sin_addr,
	 (void *)h->h_addr_list[0],
	 sizeof(myaddr.sin_addr));
  sa_len= sizeof(myaddr);
  return 0;
}

int create_mcast_socket( const char *mcast_addr_name,
			 unsigned port,
			 int ttl) {
#ifdef __WIN32__
  int fd= WSASocket( AF_INET, SOCK_DGRAM, IPPROTO_UDP,
  		     NULL, 0, 
  		     WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF);
#else
  int fd= socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP);
#endif
  if (fd == -1) {
#ifdef WIN32
    cerr << "Cannot create datagram socket: last error " 
	 << WSAGetLastError() << endl;
#else
    perror("Cannot create datagram socket");
#endif
    return -1;
  }

  sockaddr_in addr;

  hostent *hp= gethostbyname(mcast_addr_name);

  in_addr mcast_addr;
  mcast_addr.s_addr= *(long*)hp->h_addr;
  char my_host_name[128];
  if (gethostname(my_host_name,sizeof(my_host_name))) {
    perror("Cannot get host name");
    return -1;
  }
  hp= gethostbyname(my_host_name);
  in_addr ucast_addr;
  ucast_addr.s_addr= *(long*)hp->h_addr;

  int sso_res;

#ifdef __WIN32__

  sockaddr_in saddr;
  saddr.sin_family= AF_INET;
  saddr.sin_port= port;
  saddr.sin_addr= mcast_addr;

  sso_res= WSAJoinLeaf( fd, (sockaddr*) &saddr, sizeof(saddr),
			NULL, NULL, NULL, NULL,
			JL_BOTH);

#else
  ip_mreq mreq;
  mreq.imr_multiaddr= mcast_addr;
  mreq.imr_interface= ucast_addr;
#endif // __WIN32__

  int reuseaddr= 1;
  unsigned sock_buf_len= 61440; // 60 K

#ifdef __linux__
  setsockopt( fd, SOL_SOCKET,
	      SO_RCVBUF,
	      &sock_buf_len, sizeof(sock_buf_len));

  setsockopt( fd, SOL_SOCKET,
	      SO_REUSEADDR,
	      &reuseaddr, sizeof(reuseaddr));

  setsockopt( fd, SOL_IP /* IP level socket option */,
	      IP_TTL,
	      &ttl, sizeof(ttl));
  setsockopt( fd, SOL_IP /* IP level socket option */,
	      IP_MULTICAST_TTL,
	      &ttl, sizeof(ttl));
#ifndef __WIN32__	      
  sso_res= setsockopt( 	fd, SOL_IP /* IP level socket option */,
	   	   	IP_ADD_MEMBERSHIP,
	    		&mreq, sizeof(mreq));
#endif // __WIN32__

#else
  setsockopt( fd, SOL_SOCKET,
	      SO_RCVBUF,
	      (const char *) &sock_buf_len, sizeof(sock_buf_len));

  setsockopt( fd, IPPROTO_IP /* IP level socket option */,
	      IP_TTL,
	      (const char *) &ttl, sizeof(ttl));
  setsockopt( fd, IPPROTO_IP /* IP level socket option */,
	      IP_MULTICAST_TTL,
	      (const char *) &ttl, sizeof(ttl));

#ifndef __WIN32__
  sso_res= setsockopt( 	fd, IPPROTO_IP /* IP level socket option */,
	   	   	IP_ADD_MEMBERSHIP,
	    		(const char *) &mreq, sizeof(mreq));  
#endif

  setsockopt( fd, SOL_SOCKET,
	      SO_REUSEADDR,
	      (const char *) &reuseaddr, sizeof(reuseaddr));
#endif // __linux__

#ifdef __WIN32__  
  if (sso_res == -1) {
    cerr << "WSAGetLastError " << WSAGetLastError() << endl;
    return -1;
  }
  else {
    //    cerr << "setsockopt succeeded" << endl;
  }
#else
  if (sso_res) {
    perror("setsockopt failed");
    return -1;
  }
  else {
    //    cerr << "setsockopt succeeded" << endl;
  }
#endif

  sockaddr_in sa;
  sa.sin_family= AF_INET;
  sa.sin_port = htons(port);
  sa.sin_addr.s_addr = INADDR_ANY;
#ifdef __linux__
  socklen_t sa_len= sizeof(sa);
#else
  int sa_len= sizeof(sa);
#endif
  if (bind( fd, (sockaddr*) &sa, sizeof(sa))<0) {
#ifdef WIN32
    cerr << "bind failed: last error " << WSAGetLastError() << endl;
#else
    perror("bind failed");
#endif
    exit(-1);
    return -1;
  }

  return fd;
}

static sockaddr local_address;

int create_ucast_socket( unsigned port) {
  int fd= socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (fd == -1) {
    perror("Cannot create datagram socket");
    return -1;
  }
  char my_host_name[128];
  if (gethostname(my_host_name,sizeof(my_host_name))) {
    perror("Cannot get host name");
    return -1;
  }
  hostent *hp= gethostbyname(my_host_name);
  ((sockaddr_in&)local_address).sin_family= AF_INET;
  ((sockaddr_in&)local_address).sin_port= 0;
  ((sockaddr_in&)local_address).sin_addr.s_addr= *(u32*) hp->h_addr_list[0];

  sockaddr_in sa;
  sa.sin_family= AF_INET;
  sa.sin_port = htons(port);
  sa.sin_addr.s_addr= INADDR_ANY;
#ifdef __linux__
  socklen_t sa_len= sizeof(sa);
#else
  int sa_len= sizeof(sa);
#endif
  if (bind( fd, (sockaddr*) &sa, sizeof(sa))<0) {
    perror("ucast bind failed");
    close(fd);
    return -1;
  }
  return fd;
}

int TSAP::operator ==( const TSAP &tsap) const {
  if (addr_len == tsap.addr_len && id == tsap.id) {
    return !memcmp( &((sockaddr_in&)addr).sin_addr, 
		    &((sockaddr_in&)tsap.addr).sin_addr, 
		    4);
  }
  else return 0;
}

void 
MTPWeb::
update_after_join( const MTPPacket &packet) {
  r_heartbeat= packet.header.heartbeat;
  r_window= packet.header.window;
  r_retention= packet.header.retention;
  r_mcast_connection_id= ntohl(*(u32*)(&packet.data[8]));
  r_master_connection_id= packet.header.src_id;
  r_max_data_size= ntohs(*(u16*)(&packet.data[0]));
}
 
MTPPacket::
MTPPacket( const MTPPacket &src) :
  all_data( new unsigned char[MTPHeader::MTP_HEADER_SIZE+src.data_length]),
  data( all_data+MTPHeader::MTP_HEADER_SIZE),
  data_length( src.data_length) {
  
  memcpy( all_data, src.all_data, MTPHeader::MTP_HEADER_SIZE+src.data_length);
  header= src.header;
  addrfrom= src.addrfrom;
  addrfrom_len= src.addrfrom_len;
}

MTPPacket & MTPPacket::
operator =( const MTPPacket &src) {
  if (&src!=this) {
    if (all_data) delete []all_data;
    all_data= new unsigned char[MTPHeader::MTP_HEADER_SIZE+src.data_length];
    data= all_data+MTPHeader::MTP_HEADER_SIZE;
    data_length= src.data_length;
    
    memcpy( all_data, src.all_data, 
	    MTPHeader::MTP_HEADER_SIZE+src.data_length);
    header= src.header;
    addrfrom= src.addrfrom;
    addrfrom_len= src.addrfrom_len;  
  }
  return *this;
}

int MTPPacket::
send( int fd,
      sockaddr &to,
      int tolen) {

  int res= sendto( fd, (const char *) all_data, 
		   data_length+MTPHeader::MTP_HEADER_SIZE,
		   0, // no flags
		   &to, tolen);
  //  cerr << "send " << data_length+MTPHeader::MTP_HEADER_SIZE << endl;
  if (res != data_length+MTPHeader::MTP_HEADER_SIZE) {
    cerr << "send failed" << endl;
    return -1;
  }
  return 0;
}

int MTPPacket::
send_ucast_reply( int fd, unsigned port) {

  sockaddr ad= addrfrom;

  ((sockaddr_in&) ad).sin_port= htons(port);
  int res= sendto( fd, (const char *) all_data,
		   data_length+MTPHeader::MTP_HEADER_SIZE,
		   0, // no flags
		   &ad, addrfrom_len);
  if (res != data_length+MTPHeader::MTP_HEADER_SIZE) return -1;
  return 0;
}

int MTPPacket::
recv( int fd) {

  for (;;) {
    addrfrom_len= sizeof(addrfrom);

    int res= recvfrom( fd, (char *) all_data, 
		       MAX_MTP_PACKET_SIZE, 0, // no flags
		       (sockaddr*) &addrfrom, (int *) &addrfrom_len);
    //    cerr << "recv " << res << endl;
    //    cerr << "Packet received " << currtime() << endl;
    if (res<MTPHeader::MTP_HEADER_SIZE) {
      if (res<0) return -1;
      else continue; // discard packets that are too short
    }
    header.protocol_version= all_data[MTPHeader::POS_protocol_version];
    header.packet_type= all_data[MTPHeader::POS_packet_type];
    header.packet_modifier= all_data[MTPHeader::POS_packet_modifier];
    header.client_channel= all_data[MTPHeader::POS_client_channel];
    header.src_id= ntohl(*(u32*) &all_data[MTPHeader::POS_src_id]);
    header.dst_id= ntohl(*(u32*) &all_data[MTPHeader::POS_dst_id]);
    header.synchro= all_data[MTPHeader::POS_synchro];
    memcpy( header.msg_state, &all_data[MTPHeader::POS_msg_state],
	    sizeof(header.msg_state));
    header.msg_seq_number= 
      u48( ntohs(*(u16*) &all_data[MTPHeader::POS_msg_seq_number]),
	   ntohl(*(u32*) &all_data[MTPHeader::POS_msg_seq_number+2]));
    header.packet_seq_number=
      ntohs(*(u16*) &all_data[MTPHeader::POS_packet_seq_number]);
    header.heartbeat= ntohl(*(u32*) &all_data[MTPHeader::POS_heartbeat]);
    header.window= ntohs(*(u16*) &all_data[MTPHeader::POS_window]);
    header.retention= ntohs(*(u16*) &all_data[MTPHeader::POS_retention]);
    header.reply_port= ntohs(*(u16*) &all_data[MTPHeader::POS_reply_port]);
    data_length= res-MTPHeader::MTP_HEADER_SIZE;
    return 0;
  }
}

void MTPPacket::
encode_header() {
  all_data[MTPHeader::POS_protocol_version]= header.protocol_version;
  all_data[MTPHeader::POS_packet_type]= header.packet_type;
  all_data[MTPHeader::POS_packet_modifier]= header.packet_modifier;
  all_data[MTPHeader::POS_client_channel]= header.client_channel;
  *(u32*) &all_data[MTPHeader::POS_src_id]= htonl( header.src_id);
  *(u32*) &all_data[MTPHeader::POS_dst_id]= htonl( header.dst_id);
  all_data[MTPHeader::POS_synchro]= header.synchro;
  memcpy( &all_data[MTPHeader::POS_msg_state], header.msg_state,
	  sizeof(header.msg_state));
  *(u16*) &all_data[MTPHeader::POS_msg_seq_number]= 
    htons( header.msg_seq_number.get_hi());
  *(u32*) &all_data[MTPHeader::POS_msg_seq_number+2]= 
    htonl( header.msg_seq_number.get_lo());
  *(u16*) &all_data[MTPHeader::POS_packet_seq_number]=
    htons( header.packet_seq_number);
  *(u32*) &all_data[MTPHeader::POS_heartbeat]= htonl( header.heartbeat);
  *(u16*) &all_data[MTPHeader::POS_window]= htons( header.window);
  *(u16*) &all_data[MTPHeader::POS_retention]= htons( header.retention);  
  *(u16*) &all_data[MTPHeader::POS_reply_port]= htons( header.reply_port);
}

void MTPPacket::
create_join_request( unsigned char member_class,
		     const MTPWeb &web) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_JOIN;
  header.packet_modifier= MTPHeader::T_JOIN_REQUEST;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= MTP_UNKNOWN_ID;

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number.reset();
  header.packet_seq_number= 0;
  // join request must have zeroed acceptance criteria per RFC1301 3.1.1
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();

  encode_header();

  data_length= 3*4+2;
  data[0]= member_class;
  data[1]= 0; // transport class: raliable using naks
  data[2]= 0; // transport type: multiple producers, multiple consumers
  data[3]= 0; // reserved, must be zero
  *(u16*)(&data[4])= 0; // minimum throughput
  *(u16*)(&data[6])= 65535; // maximum data unit size
  *(u32*)(&data[8])= 0; // multicast connection identifier
  
}

void MTPPacket::
create_join_confirm( const MTPWeb &web) {

  // must be called for previously received "join request" packet
  // can only be sent by a master of web

  header.swap_ids();
  header.src_id= web.master_connection_id();
  header.packet_modifier= MTPHeader::T_JOIN_CONFIRM;
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number= web.sequence_number();
  header.packet_seq_number= web.packet_number();
  header.reply_port= web.reply_port();

  encode_header();

  data_length= 3*4;
  *(u16*)(&data[0])= htons( web.max_data_size());
  *(u32*)(&data[8])= htonl( web.mcast_connection_id());
}

void MTPPacket::
create_join_deny( const MTPWeb &web) {
  // must be called for previously received "join request" packet
  // can only be sent by a master of web
  
  header.swap_ids();
  header.src_id= web.master_connection_id();
  header.packet_modifier= MTPHeader::T_JOIN_DENY;
  header.reply_port= web.reply_port();
  encode_header();

  data_length= 3*4;
  *(u32*)(&data[8])= htonl( web.mcast_connection_id());
}

void MTPPacket::
create_quit_request( const MTPWeb &web,
		     u32 dst_connection_id) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_QUIT;
  header.packet_modifier= MTPHeader::T_QUIT_REQUEST;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= dst_connection_id;

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number.reset();
  header.packet_seq_number= 0;
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();

  encode_header();
  data_length= 0;
}

void MTPPacket::
create_quit_confirm( const MTPWeb &web) {
  // must be called for previously received "quit request" packet

  header.swap_ids();
  header.src_id= web.ucast_connection_id();
  header.packet_modifier= MTPHeader::T_QUIT_CONFIRM;
  header.reply_port= web.reply_port();
  encode_header();
  data_length= 0;
}

void MTPPacket::
create_is_member_request( const MTPWeb &web,
			  u32 dst_connection_id) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_IS_MEMBER;
  header.packet_modifier= MTPHeader::T_IS_MEMBER_REQUEST;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= dst_connection_id;

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number.reset();
  header.packet_seq_number= 0;
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();

  encode_header();  
  data_length= 0;
}

void MTPPacket::
create_is_member_confirm( const MTPWeb &web) {
  // must be called for previously received "is member request" packet

  header.swap_ids();
  header.src_id= web.ucast_connection_id();
  header.packet_modifier= MTPHeader::T_QUIT_CONFIRM;
  header.reply_port= web.reply_port();
  encode_header();
  data_length= 0;
}

void MTPPacket::
create_is_member_deny( const MTPWeb &web) {
  // must be called for previously received "is member request" packet

  header.swap_ids();
  header.src_id= web.ucast_connection_id();
  header.packet_modifier= MTPHeader::T_IS_MEMBER_DENY;
  header.reply_port= web.reply_port();
  encode_header();
  data_length= 0;
}

void MTPPacket::
create_token_request( const MTPWeb &web) {
  
  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_TOKEN;
  header.packet_modifier= MTPHeader::T_TOKEN_REQUEST;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= web.master_connection_id();

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number.reset();
  header.packet_seq_number= 0;
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();
  encode_header();    
  data_length= 0;
}

void MTPPacket::
create_token_confirm( const MTPWeb &web) {
  // must be called for previously received "token request" packet
  // can only be called be a master of web

  header.swap_ids();
  header.src_id= web.master_connection_id();
  header.packet_modifier= MTPHeader::T_TOKEN_CONFIRM;
  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number= web.sequence_number();
  header.packet_seq_number= web.packet_number();
  //  cerr << "seq_number " << web.sequence_number() << endl;
  header.reply_port= web.reply_port();
  encode_header();
  data_length= 0;
}

void MTPPacket::
create_empty_packet( const MTPWeb &web,
		     u8 modifier 
		     // T_EMPTY_DALLY, T_EMPTY_CANCEL, T_EMPTY_HIBERNATE
		     ) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_EMPTY;
  header.packet_modifier= modifier;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= web.mcast_connection_id();

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number= web.sequence_number();
  header.packet_seq_number= web.packet_number();
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();
  encode_header();    
  data_length= 0;
}

void MTPPacket::
create_data_packet( const MTPWeb &web,
		    u8 modifier,  
 		    // T_DATA_DATA, T_DATA_EOM, T_DATA_EOW
		    unsigned a_data_length) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_DATA;
  header.packet_modifier= modifier;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= web.mcast_connection_id();

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number= web.sequence_number();
  header.packet_seq_number= web.packet_number();
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();
  encode_header();    
  data_length= a_data_length;
}

unsigned MTPPacket::  // returns number of nak request ranges consumed
create_nak_request( const MTPWeb &web,
		    u32 producer_connection_id,
		    XSTD::vector<NAKRange> &ranges) {

  header.protocol_version= MTP_CURRENT_VERSION;
  header.packet_type= MTPHeader::T_NAK;
  header.packet_modifier= MTPHeader::T_NAK_REQUEST;
  header.client_channel= 0;
  header.src_id= web.ucast_connection_id();
  header.dst_id= producer_connection_id;

  header.synchro= 0;
  memset( header.msg_state, 0, sizeof(header.msg_state));
  header.msg_seq_number.reset();
  header.packet_seq_number= 0;
  
  header.heartbeat= web.heartbeat();
  header.window= web.window();
  header.retention= web.retention();
  header.reply_port= web.reply_port();
  encode_header();
  unsigned ret;
  unsigned max_nak_ranges= web.max_data_size()/(16);
  
  typedef XSTD::vector<NAKRange> NAKVector;

  NAKVector::const_iterator it( ranges.begin());
  for (ret=0; 
       ret<max_nak_ranges && it != ranges.end(); 
       ret++, it++) {
    *(u16*) &data[ret*16]= htons((*it).min_packet);
    *(u16*) &data[ret*16+2]= htons((*it).min_message.get_hi());
    *(u32*) &data[ret*16+4]= htonl((*it).min_message.get_lo());

    *(u16*) &data[ret*16+8]= htons((*it).max_packet);
    *(u16*) &data[ret*16+10]= htons((*it).max_message.get_hi());
    *(u32*) &data[ret*16+12]= htonl((*it).max_message.get_lo());
  }
  data_length= ret*16;
  return ret;
}

void MTPPacket::
create_nak_deny( const MTPWeb &web) {
  // must be called for previously received "nak request" packet

  header.swap_ids();
  header.src_id= web.ucast_connection_id();
  header.packet_modifier= MTPHeader::T_NAK_DENY;
  header.reply_port= web.reply_port();
  encode_header();
  data_length= 0;
}

PacketMgr::PacketMgr( const char *mcast_address,
		      unsigned _mtplistenport,
		      int _ttl) :
  PacketEndpoint( _mtplistenport),
  free_handlers( NULL),
  busy_handlers( NULL),
  address( new char[strlen(mcast_address)+1]),
  listen_port( _mtplistenport)
{
  strcpy( address, mcast_address);

  trace.start(1000)
    << "PacketMgr: ListenPort " << _mtplistenport
    << " @ " << mcast_address << '\n';  
  
  fd= create_mcast_socket( mcast_address,
			   mtplistenport,
			   _ttl);
  for (reply_port= mtplistenport+1; ; reply_port++)
    if ((ucast_fd= create_ucast_socket(reply_port)) != -1) break;

  out_fd= socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP);

  if (resolve_host( mcast_address, 
		    send_addr, send_addr_len, mtplistenport)) {
    perror("Invalid address");
    exit(1);
  }
  ipaddr= ((sockaddr_in &)send_addr).sin_addr.s_addr;
    
  if (fd!=-1) {
    unsigned init_mgrs= 3;
    for (unsigned i=0; i<init_mgrs; i++)
      free_handlers= new INPHandler( *this, free_handlers, NULL, i);
  }
}

PacketMgr::~PacketMgr() {
  delete []address;
}

void PacketMgr::send_mcast( MTPPacket &packet) {
  // a destination connection identifier is either 0 (unknown id)
  // or a multicast connection identifier gleaned from join_confirm packet
  //
  packet.send( out_fd, send_addr, send_addr_len);
}

void PacketMgr::send_ucast_reply( MTPPacket &packet, u16 port) {
  packet.send_ucast_reply( out_fd, port);
}

void PacketMgr::register_handler( IncomingPacketHandler &hnd,
				  u32 ucast_connection_id,
				  u32 my_id) {

  hmap.insert(map_type::value_type( ucast_connection_id, 
				    MapItem(&hnd, my_id)));
}

int PacketMgr::unregister_handler( IncomingPacketHandler &hnd,
				   u32 ucast_connection_id) {

  XSTD::pair<map_type::iterator, map_type::iterator> p =
    hmap.equal_range(ucast_connection_id);
  for (map_type::iterator i = p.first; i != p.second; ++i)
    if ((*i).second.iph == &hnd) {
      hmap.erase(i);
      return 1;
    }
  return 0;  
}

void PacketMgr::INPHandler::handler( void *ptr) {
  INPHandler &th= *(INPHandler*)ptr;

  for (;;) {
    ((sockaddr_in&) th.packet.addrfrom).sin_port= 0;
    int res= (th.no) ? th.packet.recv( th.mgr.get_fd()) :
      th.packet.recv( th.mgr.get_ucast_fd());

    int is_foreign= 
      !(TSAP(th.packet.addrfrom, th.packet.addrfrom_len, 0) ==
	TSAP(local_address, sizeof(local_address), 0));
    if (res != -1) {
      XSTD::pair<map_type::iterator, map_type::iterator> p =
	th.mgr.hmap.equal_range(th.packet.header.dst_id);
      for (map_type::iterator i = p.first; i != p.second; ++i) {
	if (is_foreign || th.packet.header.src_id != (*i).second.my_connection_id)
	  (*i).second.iph->handle( th.packet);
      }
    }
  }
}

//-----------------------------------------------------------PacketMgrFactory--
//
int PacketMgrFactory::ttl;
unsigned PacketMgrFactory::listen_port;
sockaddr PacketMgrFactory::send_addr;
unsigned PacketMgrFactory::send_addr_len;
unsigned PacketMgrFactory::address_range;

void PacketMgrFactory::init( int &argc, char ** &argv) {

#ifdef __WIN32__
  WORD wVersionRequested;
  WSADATA wsaData;
  wsaData.lpVendorInfo= NULL;
  int err;
  
  wVersionRequested = MAKEWORD( 2, 0 );
  
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) {
    // Tell the user that we couldn't find a usable
    // WinSock DLL.
    cerr << "This application requires Winsock version 2.0 (or higher)" << endl;
    exit(-1);
  }

  // Confirm that the WinSock DLL supports 2.0.
  // Note that if the DLL supports versions greater
  // than 2.0 in addition to 2.0, it will still return
  // 2.0 in wVersion since that is the version we
  // requested.
 
  if ( LOBYTE( wsaData.wVersion ) != 2 ||
       HIBYTE( wsaData.wVersion ) != 0 ) {
    // Tell the user that we couldn't find a usable
    // WinSock DLL.
    cerr << "This application requires Winsock version 2.0 (or higher)" << endl;
    WSACleanup( );
    exit(-1);
  }
 
  // The WinSock DLL is acceptable. Proceed.

  cerr << "WINSOCK initialization" << '\n'
       << wsaData.szDescription << '\n'
       << (wsaData.lpVendorInfo ? wsaData.lpVendorInfo : "") 
       << endl;
#endif
  
  listen_port= 4096;
  ttl= -1; // default
  address_range= 32; // 32 multicast addresses 
  char *mtpaddress= new char[ strlen(MTP_PERMANENT_ADDR)+1];
  strcpy( mtpaddress, MTP_PERMANENT_ADDR);

  char *env= getenv("MTP_OPTIONS");
  if (env) {
    istrstream istr(env);
    char optname[100];
    while (istr >> optname) {
      unsigned val;
      if (!strcmp(optname, "-MTPListenPort")) {
	istr >> val;
	listen_port= val;
      }
      if (!strcmp(optname, "-MTPAddress")) {
	istr >> optname;
	delete [] mtpaddress;
	strcpy( mtpaddress, optname);
      }
      if (!strcmp(optname, "-MTPRange")) {
	istr >> val;
	address_range= val;
      }
      if (!strcmp(optname, "-MTPTTL")) {
	istr >> ttl;
      }
    }
  }
  
  int org_len= 0;
  for (char **x_ptr= argv; *x_ptr; x_ptr++)
    org_len++;

  char ** nargv= (char**) malloc((org_len+1)*sizeof(char*));
  argc= 0;
  for (char **n_ptr= argv; *n_ptr; n_ptr++) {
    if (!strcmp("-MTPListenPort",*n_ptr) && *(n_ptr+1)) {
      listen_port= atoi( *(++n_ptr));
      ++n_ptr;
      continue;
    }
    if (!strcmp("-MTPAddress", *n_ptr) && *(n_ptr+1)) {
      delete []mtpaddress;
      ++n_ptr;
      mtpaddress= new char[ strlen(*n_ptr)+1];
      strcpy( mtpaddress, *n_ptr);
      ++n_ptr;
    }
    if (!strcmp("-MTPRange", *n_ptr) && *(n_ptr+1)) {
      address_range= atoi( *(++n_ptr));
      ++n_ptr;
      continue;
    }
    if (!strcmp("-MTPTTL", *n_ptr) && *(n_ptr+1)) {
      ttl= atoi( *(++n_ptr));
      ++n_ptr;
      continue;
    }    
    nargv[argc++]= *n_ptr;
  }
  nargv[argc]= NULL;
  resolve_host( mtpaddress, send_addr, send_addr_len, 0);
  trace.start(1000)
    << "PacketMgrFactory: ListenPort " << listen_port
    << " @ " << mtpaddress << '\n';
  
  argv= nargv;
}

// this method is used by multicast sink in order to create a communication
// endpoint having an endpoint description (address and port number)
//
PacketMgr *PacketMgrFactory::
find_or_create( const char *endpoint_id) {

  omni_mutex_lock oml(guard);
  char mcast_addr[100];
  unsigned my_listen_port;

  //  cerr << "find or create " << endpoint_id << endl;
  sscanf(endpoint_id, "%s%d", mcast_addr, &my_listen_port);
  PacketEndpoint pe( my_listen_port);
  sockaddr sa;
  unsigned sa_len;
  resolve_host( mcast_addr, sa, sa_len, 0);
  pe.ipaddr= ((sockaddr_in&) sa).sin_addr.s_addr;

  typedef XSTD::set<PacketEndpoint*,cmpmgr> Mgrs;
  Mgrs::iterator fr= mgrs.find( &pe);
  if (fr != mgrs.end()) return (PacketMgr*) (*fr);
  PacketMgr *pm= new PacketMgr( mcast_addr,
				my_listen_port,
				ttl);
  mgrs.insert( pm);
  return pm;
}

// creates or reuses a packet manager
//
PacketMgr *PacketMgrFactory::
assign_channel() {

  omni_mutex_lock oml(guard);

  u32 base_addr= ((sockaddr_in&) send_addr).sin_addr.s_addr;
  ((u8*)(&base_addr))[3]+= (addresses_in_use++);

  PacketEndpoint pe( listen_port);
  pe.ipaddr= base_addr;
  typedef XSTD::set<PacketEndpoint*,cmpmgr> Mgrs;
  Mgrs::iterator fr= mgrs.find( &pe);

  if (addresses_in_use >= address_range) {
    // reuse address
    //
    return (PacketMgr*) (*fr);
  }
  else if (fr == mgrs.end()) {
    char my_addr[100];
    sprintf(my_addr,"%u.%u.%u.%u", 
	    ((u8*)(&base_addr))[0], ((u8*)(&base_addr))[1],
	    ((u8*)(&base_addr))[2], ((u8*)(&base_addr))[3]);
	    
    PacketMgr *pm= new PacketMgr( my_addr, listen_port, ttl);
    mgrs.insert( pm);
    return pm;
  }
}

#else
int mcast_dummy; // it keeps the SparcCompiler 4.2 linker happy

#endif // NO_MCAST

//
