// $Id: mtpsink.hh,v 1.1 1998/11/11 02:22:08 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtpsink.hh,v $
//
// $Log: mtpsink.hh,v $
// Revision 1.1  1998/11/11 02:22:08  mencda
// Initial revision
//
//
//

#ifndef __MTPSINK_HH__
#define __MTPSINK_HH__

#include "mtp.hh"
#include "mcast.hh"
#include "utils.hh"

#include <map>
#include <deque>


class MessageReady {
public:
  unsigned len,rd_ptr;
  unsigned char *buf;
  
  MessageReady() : buf(NULL), len(0), rd_ptr(0) {}
  MessageReady( const MessageReady &src) : buf(NULL), len(0), rd_ptr(0) {
    if (src.buf) {
      buf= new unsigned char[len= src.len];
      rd_ptr= 0;
      memcpy( buf, src.buf, len);
    }
  }
  MessageReady & operator =( const MessageReady &src) {
    if (&src != this) {
      if (buf) delete []buf;
      buf= NULL;
      len= src.len;
      rd_ptr= 0;
      if (src.buf) {
	buf= new unsigned char[len= src.len];
	memcpy( buf, src.buf, len);
      }
    }
    return *this;
  }
  ~MessageReady() { delete []buf; }
  
  void prepare( unsigned a_len) {
    buf= new unsigned char[len= a_len];
  }
  
  int operator <( const MessageReady &) const;
  int operator ==(const MessageReady &) const;
};


// Message assemble and retransimission
//
class MessageAssemble {
public:
  MessageAssemble() : last_accepted(0), accepted(0) {}
  
  u16 last_accepted;
  int accepted;
  typedef XSTD::map<u16,MTPPacket,XSTD::less<u16> > MapAssemble;
  MapAssemble packets;
};


class MTPSink : private IncomingPacketHandler {

  typedef XSTD::map<u48,MessageAssemble,XSTD::less<u48> > Msgs;
  Msgs assemble;
  u48 last_message_accepted;
  void send_naks( MTPPacket &packet);

  // Message buffering
  //
  typedef XSTD::deque<MessageReady> GBuf;
  GBuf grow_buffer;

  PacketMgr &packetmgr;
  omni_mutex guard;
  omni_condition cond_wait, cond_wait_for_data;
  u32 my_ucast_id;
  MTPWeb web;
  TSAP producer;
  TSAP master;
  MTPPacket resp_packet;
  MTPPacket out_packet;
  enum { STATE_INIT, STATE_FAILED, STATE_ESTABLISHED } state;

  void init();
  virtual void handle( MTPPacket &packet);
  void handle_data_packet( MTPPacket &packet, int dally);
public:

  MTPSink( PacketMgr &a_packetmgr) : 
    cond_wait(&guard), cond_wait_for_data(&guard), packetmgr(a_packetmgr) {

    producer.id= 0;
    init();
  }

  ~MTPSink();

  void read( void *buf, unsigned long buflen);

  DISABLE_COPY_AND_ASSIGN(MTPSink)
};

#endif // __MTPSINK_HH__

//
