// $Id: mtpsource.cc,v 1.1 1998/11/06 03:25:01 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtpsource.cc,v $
//
// $Log: mtpsource.cc,v $
// Revision 1.1  1998/11/06 03:25:01  mencda
// Initial revision
//
//

#ifndef NO_MCAST

#include "mtpsource.hh"

#include <iostream.h>

MTPSource::~MTPSource() {
  delete []out_buffer;
  packetmgr.unregister_handler( *this, my_ucast_id);
}

void MTPSource::send_cancel_packet() {
  out_packet.create_empty_packet( web, MTPHeader::T_EMPTY_CANCEL);
  out_packet.addrfrom= master.addr;
  out_packet.addrfrom_len= master.addr_len;
  packetmgr.send_ucast_reply( out_packet, master.reply_port);
}

void MTPSource::send_token_request() {
  out_packet.create_token_request( web);
  out_packet.addrfrom= master.addr;
  out_packet.addrfrom_len= master.addr_len;
  packetmgr.send_ucast_reply( out_packet, master.reply_port);
  //  cerr << "send_token_request to port " << master.reply_port << endl;
}

void MTPSource::build_packet( u8 data_modifier,
			      unsigned long len) {
  
  web.assign_packet_number();
  out_packet.create_data_packet( web,
				 data_modifier, len);
  unsigned long rd_pos= 
    (out_buffer_pos+out_buffer_len-out_buffer_usage) % out_buffer_len;
  unsigned flen= out_buffer_len-rd_pos;
  flen= flen>len ? len : flen;
  memcpy( out_packet.data, out_buffer+rd_pos, flen);
  out_buffer_usage-=flen;
  len-=flen;
  if (len)
    memcpy( out_packet.data+flen, out_buffer, len);
  out_buffer_usage-=len;

}

void MTPSource::send_data_packet() {
  trace.start(10000) 
    << "SEND_DATA " << out_packet.header.msg_seq_number << '.'
    << out_packet.header.packet_seq_number << '\n';
  packetmgr.send_mcast( out_packet);
  if (retention_buffer.size() > 50) retention_buffer.pop_front();
  retention_buffer.push_back( out_packet);
}

void MTPSource::reply_nak_deny( MTPPacket &packet) {
  u16 reply_port= packet.header.reply_port;
  packet.create_nak_deny( web);
  packetmgr.send_ucast_reply( packet, reply_port);
}

void MTPSource::handle_nak( MTPPacket &packet) {
  //  cerr << "handle_nak " << packet.data_length << endl;
  if (retention_buffer.empty()) {
    reply_nak_deny( packet);
    return;
  }
  
  unsigned len= packet.data_length/16;
  if (len==0) return;
  
  RetentionBuffer::iterator it= retention_buffer.begin();

  for (unsigned i=0; i<len; i++) {
    u16 min_pck= ntohs(*((u16*)&packet.data[0+i*16]));
    u48 min_seq( ntohs(*((u16*)&packet.data[2+i*16])),
		 ntohl(*((u32*)&packet.data[4+i*16])));

    u16 max_pck= ntohs(*((u16*)&packet.data[8+i*16]));
    u48 max_seq( ntohs(*((u16*)&packet.data[10+i*16])),
		 ntohl(*((u32*)&packet.data[12+i*16])));

    trace.start(10000)
      << "NAK [ " << min_seq << ", "
      << min_pck << "] -> [ " << max_seq << ", "
      << max_pck << "]" << '\n';
    
    
    int reversed= 0;
    if ( min_seq > max_seq)
      reversed= 1;
    else if ( min_seq == max_seq && min_pck > max_pck)
      reversed= 1;

    // let's find the beginning of missing packet sequence
    //
    while ( it!= retention_buffer.end() && 
	    !(min_seq == (*it).packet.header.msg_seq_number &&
	      min_pck == (*it).packet.header.packet_seq_number)) {
      //      cerr << (*it).header.msg_seq_number << ','
      //	   << (*it).header.packet_seq_number << endl;
      it++;
    }
    if (it == retention_buffer.end()) {
      trace.start(10000)
	<< "Start of NAK request not found in retention buffer" << '\n';
      reply_nak_deny( packet);
      return;
    }

    // we have found the beginning of packet sequence
    //
    PacketID minid( min_seq,min_pck);
    PacketID maxid( max_seq,max_pck);
    PacketID currid= PacketID((*it).packet.header.msg_seq_number,
			      (*it).packet.header.packet_seq_number);

    while ( ((!reversed && currid >= minid && currid < maxid) ||
	     (reversed && (currid >= minid || currid < maxid)))) {

      Time ct;
      omni_thread::get_time( &ct.sec, &ct.nsec);
      if ( ct.diff( (*it).time) < Time(0,50000000 /* 50 ms */)) {
	// it to soon for a new retransmission
      }
      else {
	trace.start(10000)
	  << "RETRANSMISSION " << (*it).packet.header.msg_seq_number << '.'
	  << (*it).packet.header.packet_seq_number << '\n';
	packetmgr.send_mcast( (*it).packet);
	(*it).time= ct;
      }
      it++;
      if (it == retention_buffer.end()) break;
      currid= PacketID((*it).packet.header.msg_seq_number,
		       (*it).packet.header.packet_seq_number);
    }
  }
}

void MTPSource::got_token( MTPPacket &packet) {
  //  cerr << "MTPSource::got_token" << endl;

  if (!out_buffer_usage) {
    send_cancel_packet();
    has_token= 0;
    return;
  }
  has_token= 1;
  web.set_sequence_number( packet.header.msg_seq_number);
  for (;;) {
    while (out_buffer_usage >= web.max_data_size()) {
      build_packet( MTPHeader::T_DATA_DATA, web.max_data_size());
      send_data_packet();
      cond_full.signal(); // wake up threads waiting for a buffer
    }
    if (eom) {
      //      cerr << "MTPSource::got_token eom" << endl;
      // now buffer usage may only drop, as sending client is performing
      // flush();
      build_packet( MTPHeader::T_DATA_EOM, out_buffer_usage);
      send_data_packet();
      cond_full.signal();
      break;
    }
    //    cerr << "MTPSource::cond_sender wait" << endl;
    cond_sender.wait();
  }
  has_token= 0;
  cond_full_eom.signal();
}


void MTPSource::flush() {
  omni_mutex_lock oml(guard);
  //  cerr << "flush (first_req= " << first_req << ")" << endl;
  eom= 1;
  cond_sender.signal();
  cond_trans.signal();
  while (out_buffer_usage)
    cond_full_eom.wait();
  if (has_token) send_cancel_packet();
  first_req= 1;
  eom= 0;
  cond_wait.signal();
}

void MTPSource::message_data( void *data, unsigned long len) {
  omni_mutex_lock oml(guard);

  unsigned char *xdata= (unsigned char *) data;
  do {
    unsigned long avail= out_buffer_len-out_buffer_usage;
    if (avail > 0) {
      unsigned long flen= (avail>len) ? len : avail;
      unsigned long len1= out_buffer_len-out_buffer_pos;
      len1= (flen>len1) ? len1 : flen;
      unsigned long len2= flen-len1;
      memcpy( &out_buffer[out_buffer_pos], xdata, len1);
      if (len2) {
	memcpy( out_buffer, xdata+len1, len2);
	out_buffer_pos= len2;
      }
      else {
	out_buffer_pos= (out_buffer_pos + len1) % out_buffer_len;
      }
      len-= (len1+len2);
      xdata+= (len1+len2);
      out_buffer_usage+= flen;
      cond_trans.signal();
      cond_sender.signal();
    }
    else {
      cond_trans.signal();
      cond_sender.signal();
      if (out_buffer_len-out_buffer_usage == 0) cond_full.wait();
    }
  } while (len);
}

void MTPSource::init() {
  my_ucast_id= mtpidmgr.assign_id();
  trace.start(1000) << "my_ucast_id " << my_ucast_id << '\n';
  web= MTPWeb( 300, // msec -heartbeat 
	       15,  // packets -window
	       10,  // retention
	       MAX_MTP_PACKET_SIZE, my_ucast_id, packetmgr.get_ucast_port());
  MTPPacket tryjoin;

  omni_mutex_lock oml(guard);
  packetmgr.register_handler( *this, my_ucast_id, my_ucast_id);
  //  cerr << "trying" << endl;
  for (int try_client=0; try_client<10; try_client++) {
    tryjoin.create_join_request( MTPHeader::MEMBER_PRODUCER,
				 web);
    packetmgr.send_mcast( tryjoin);
    unsigned long sec,nsec;
    omni_thread::get_time( &sec,&nsec, 0, 30000000 /* 30 msec */);
    cond_wait.timedwait( sec, nsec);
    if (state != STATE_INIT) break;
  }
  if (state == STATE_INIT) {
    trace.start(1000) << "Failed to find web master" << '\n';
    state= STATE_FAILED;
    return;
  }
  trace.start(1000) << "MTPSource initialized, waiting for messages" << '\n';
}

void MTPSource::my_run() {
  omni_mutex_lock oml(guard);
  
  unsigned long sec,nsec;
  for (;;) {
    int was_first_req= first_req;
    if (first_req) {
      //      cerr << "cond_trans wait" << endl;
      cond_trans.wait();
    }
    else {
      omni_thread::get_time( &sec,&nsec, 0, web.heartbeat()*1000000L);
      if (cond_wait.timedwait(sec,nsec)) { // condition was signalled
	if (!eom) continue;
      }
      else if (first_req) continue;
    }
    if (!has_token) {
      // cerr << "token request" << endl;
      send_token_request();
      first_req= 0;
    }
  }
}

void MTPSource::handle( MTPPacket &packet) {
  omni_mutex_lock oml(guard);

  u16 reply_port;
  reply_port= packet.header.reply_port;

  switch (packet.header.packet_type) {
  case MTPHeader::T_JOIN:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_JOIN_CONFIRM:
      if (state == STATE_INIT) {
	web.update_after_join( packet);
	master.addr= packet.addrfrom;
	master.addr_len= packet.addrfrom_len;
	master.reply_port= packet.header.reply_port;
	state= STATE_ESTABLISHED;
	cond_init.signal();
      }
      break;
    case MTPHeader::T_JOIN_DENY:
      if (state == STATE_INIT) state= STATE_FAILED;
      break;
    }
  case MTPHeader::T_QUIT:
    // implement: quitting a web on demand
    break;
  case MTPHeader::T_IS_MEMBER:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_IS_MEMBER_REQUEST:
      // all I can do is to confirm the membership of myself
      if (packet.header.dst_id == my_ucast_id)
	packet.create_is_member_confirm( web);
      else
	packet.create_is_member_deny( web);
      packetmgr.send_ucast_reply( packet, reply_port);
      break;
    }
    break;
  case MTPHeader::T_TOKEN:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_TOKEN_CONFIRM:
      got_token( packet);
      break;   
    }
    break;
  case MTPHeader::T_NAK:
    if (packet.header.packet_modifier == MTPHeader::T_NAK_REQUEST) {
      handle_nak( packet);
    }
    break;
  }
}

#endif // NO_MCAST

//
