// $Id: mtpwebd.cc,v 1.1 1998/11/03 18:05:09 mencda Exp mencda $
//
// $Source: /home/mencda/thesis/mtp/mtpwebd.cc,v $
//
// $Log: mtpwebd.cc,v $
// Revision 1.1  1998/11/03 18:05:09  mencda
// Initial revision
//
//

// MTP web daemon

#ifndef NO_MCAST

#include "mtp.hh"
#ifdef __WIN32__
#include <windows.h>
#else
#include <netinet/in.h>
#endif
#include "mtpwebd.hh"
#include <list>
#include <iostream.h>

// Join request
// 
struct JoinReq {
  TSAP tsap;        
  // TSAP of the originator of request
  //

  u8 member_class;
  // requested member class
  //

  int operator < (const JoinReq &) const { return 0; }
  int operator > (const JoinReq &) const { return 0; }
  int operator == (const JoinReq &) const { return 0; }
  int operator != (const JoinReq &) const { return 0; }
};

class RunMaster : public IncomingPacketHandler {
  omni_mutex guard;
  omni_condition cond;
  omni_condition cond_init;
  omni_condition cond_hibernate;
  u32 my_ucast_id;
  MTPWeb web;
  XSTD::list<JoinReq> join_reqs;
  XSTD::list<TSAP> token_reqs;
  TSAP token_owner;
  MTPPacket resp_packet;
  PacketMgr &packetmgr;
  int hibernate_counter;

  enum { TOKEN_IDLE, TOKEN_PENDING, TOKEN_BUSY } token_state;
  enum { STATE_INIT, STATE_TRY_MASTER, STATE_FAILED, STATE_ESTABLISHED } state;

  static void static_run( void *);
public:
  RunMaster( PacketMgr &a_packetmgr) : 
    cond( &guard), cond_init( &guard), cond_hibernate( &guard),
    state( STATE_INIT), token_state( TOKEN_IDLE),
    packetmgr( a_packetmgr), hibernate_counter(0) {}

  int initialize();
  void run();
  virtual void handle( MTPPacket &packet);
  
  void got_token( const MTPPacket &packet);
  void request_token( const MTPPacket &packet);
  void request_join( const MTPPacket &packet);

  void send_join_confirm();
  void send_token();
  void send_token_to( const TSAP &);
};

void RunMaster::got_token( const MTPPacket &packet) {

  // Check for retransmissions

  if (packet.header.packet_type != MTPHeader::T_TOKEN) {
    if (web.sequence_number() > packet.header.msg_seq_number) return;
    if (web.sequence_number() == packet.header.msg_seq_number) 
      if (web.packet_number() >= packet.header.packet_seq_number) return;
    
 
    web.set_sequence_number( packet.header.msg_seq_number);
    web.set_packet_number( packet.header.packet_seq_number);
  }
  
  trace.start(10000) << "master:: got_token" << '\n';
  token_state= TOKEN_IDLE;

  // any join requests pending ?
  send_join_confirm();
  // any requests waiting for token ?
  send_token();
  //
  if (token_state== TOKEN_IDLE) {
    hibernate_counter= 0;
    cond_hibernate.signal();
  }
}

void RunMaster::send_token_to( const TSAP &tsap) {
  resp_packet.create_token_request( web);
  resp_packet.addrfrom= tsap.addr;
  resp_packet.addrfrom_len= tsap.addr_len;
  resp_packet.header.src_id= tsap.id;
  resp_packet.create_token_confirm( web);
  token_owner= tsap;
  packetmgr.send_ucast_reply( resp_packet, tsap.reply_port);
}

void RunMaster::send_token() {
  if (!token_reqs.empty()) {  // precondition
    //    cerr << "master: send_token" << endl;
    token_state= TOKEN_PENDING;
    web.assign_sequence_number();
    send_token_to( token_reqs.front());
    token_reqs.pop_front();
  }
}

void RunMaster::send_join_confirm() {
  while (!join_reqs.empty()) { // precondition
    const JoinReq &jr= join_reqs.front();
    resp_packet.create_join_request( jr.member_class, web);
    resp_packet.addrfrom= jr.tsap.addr;
    resp_packet.addrfrom_len= jr.tsap.addr_len;
    resp_packet.header.src_id= jr.tsap.id;
    resp_packet.header.reply_port= jr.tsap.reply_port;
    resp_packet.create_join_confirm( web);
    packetmgr.send_ucast_reply( resp_packet, jr.tsap.reply_port);
    join_reqs.pop_front();
  }

  // postcondition: join_reqs.empty()
}

void RunMaster::request_token( const MTPPacket &packet) {
  TSAP tsap;
  tsap.addr= packet.addrfrom;
  tsap.addr_len= packet.addrfrom_len;
  tsap.id= packet.header.src_id;
  tsap.reply_port= packet.header.reply_port;
  token_reqs.push_back( tsap);
  //  cerr << "master: request_token" << endl;
  if (token_state == TOKEN_IDLE) send_token();
}

void RunMaster::request_join( const MTPPacket &packet) {
  JoinReq jr;
  jr.tsap.addr= packet.addrfrom;
  jr.tsap.addr_len= packet.addrfrom_len;
  jr.tsap.id= packet.header.src_id;
  jr.tsap.reply_port= packet.header.reply_port;
  jr.member_class= packet.data[0];
  join_reqs.push_back( jr);
  if (token_state == TOKEN_IDLE) send_join_confirm();
}

void RunMaster::static_run( void *args) {
  ((RunMaster*)args)->run();
}

void RunMaster::run() {
  token_owner.id= my_ucast_id= mtpidmgr.assign_id();
  trace.start(1000) << "my_ucast_id " << my_ucast_id << '\n';
  web= MTPWeb( 300, // msec -heartbeat 
	       15, // packets -window
	       10, // retention
	       2800, 
	       mtpidmgr.assign_id() | 1, // multicast connection id
	       my_ucast_id, // master connection id
	       packetmgr.get_ucast_port()
	       );
  MTPPacket tryjoin;

  omni_mutex_lock oml(guard);
  packetmgr.register_handler( *this, 0, my_ucast_id);
  packetmgr.register_handler( *this, my_ucast_id, my_ucast_id);
  packetmgr.register_handler( *this, 
			       web.mcast_connection_id(),
			       my_ucast_id);

  for (int try_client=0; try_client<10; try_client++) {
    tryjoin.create_join_request( MTPHeader::MEMBER_CONSUMER,
				web);
    packetmgr.send_mcast( tryjoin);
    unsigned long sec,nsec;
    omni_thread::get_time( &sec,&nsec, 0, 10000000 /* 10 msec */);
    cond.timedwait( sec, nsec);
    if (state == STATE_FAILED) {
      cond_init.signal();
      return;
    }
  }

  state= STATE_TRY_MASTER;
  
  for (int try_master=0; try_master<10; try_master++) {
    tryjoin.create_join_request( MTPHeader::MEMBER_MASTER,
				 web);
    packetmgr.send_mcast( tryjoin);
    unsigned long sec,nsec;
    omni_thread::get_time( &sec,&nsec, 0, 10000000 /* 10 msec */);
    cond.timedwait( sec, nsec);
    if (state == STATE_FAILED) {
      cond_init.signal();
      return;
    }
  }

  state= STATE_ESTABLISHED;
  cond_init.signal();
  trace.start(1000) << "Master established" << '\n';

  for (;;) {
    cond_hibernate.wait();
    if (hibernate_counter==0)
      do {
	web.assign_packet_number();
	hibernate_counter= 1;
	for (;hibernate_counter!= 0 && hibernate_counter<6; ) {
	  unsigned long sec,nsec;
	  omni_thread::get_time( &sec,&nsec, 0, 1000000L * web.heartbeat());
	  cond_hibernate.timedwait( sec, nsec);
	
	  if (token_state != TOKEN_IDLE) {
	    hibernate_counter=1000;
	    continue;
	  }

	  if (hibernate_counter) {
	    resp_packet.create_empty_packet( web, 
					     MTPHeader::T_EMPTY_DALLY);
	    packetmgr.send_mcast( resp_packet);
	    
	    hibernate_counter++;
	  }
	}
      } while ( hibernate_counter == 0);
  }
}

void RunMaster::handle( MTPPacket &packet) {
  omni_mutex_lock oml(guard);

  u16 reply_port;
  reply_port= packet.header.reply_port;

  switch (packet.header.packet_type) {
  case MTPHeader::T_JOIN:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_JOIN_REQUEST:
      switch (packet.data[0]) { // member class
      case MTPHeader::MEMBER_MASTER: {
	trace.start(1000) << "recv: join_request master" << '\n';
	u8* recv_addr= (u8*) &(((sockaddr_in&) packet.addrfrom).sin_addr);
	packet.create_join_deny( web);
	trace.start(10000) << "send deny" << '\n';
	packetmgr.send_ucast_reply( packet, reply_port);
	break;
      }
      case MTPHeader::MEMBER_CONSUMER:
      case MTPHeader::MEMBER_PRODUCER:
	if (state != STATE_ESTABLISHED) {
	  cerr << (unsigned) state << endl;
	  trace.start(1000) << "recv: join_request producer/consumer (ign)" << '\n';
	  break;
	}
	trace.start(1000) << "recv: join_request producer/consumer" << '\n';
	request_join( packet);
	break;
      }
      break;
    case MTPHeader::T_JOIN_DENY:
      if (state == STATE_INIT || state == STATE_TRY_MASTER) {
	// Another master is active
	u32 base_addr= ((sockaddr_in&) packet.addrfrom).sin_addr.s_addr;
	u8 *badr= (u8*)&base_addr;
	cerr << "Another master active at "
	     << badr[0] << '.' << badr[1] << '.' 
	     << badr[2] << '.' << badr[3] << endl;
	state= STATE_FAILED;
	cond.signal();
      }
      break;
    case MTPHeader::T_JOIN_CONFIRM:
      //      cerr << "recv: join_confirm" << endl;
      if (state == STATE_INIT) {
	// Another master active
	state= STATE_FAILED;
	cond.signal();
      }
      break;
    }
  case MTPHeader::T_QUIT:
    break;
  case MTPHeader::T_IS_MEMBER:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_IS_MEMBER_REQUEST:
      // all I can do is to confirm the membership of myself
      if (packet.header.dst_id == my_ucast_id)
	packet.create_is_member_confirm( web);
      else
	packet.create_is_member_deny( web);
      packetmgr.send_ucast_reply( packet, reply_port);
      break;
    }
    break;
  case MTPHeader::T_TOKEN:
    //    cerr << "master: Token request" << endl;
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_TOKEN_REQUEST:
      if (state != STATE_ESTABLISHED) break;
      if (token_state == TOKEN_PENDING && 
	  token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id)) {
	// sending client might have missed token_confirm packet
	send_token_to( token_owner);
	break;
      }
      if (token_state == TOKEN_BUSY && 
	  token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id)) {
	//	cerr << "the master has missed an EOM packet" << endl;
	// the master has missed an EOM packet
	got_token(packet);
	request_token(packet);
	break;
      }
      request_token( packet);
    }
    break;
  case MTPHeader::T_EMPTY:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_EMPTY_DALLY:
      if (token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id) &&
	  packet.header.msg_seq_number == web.sequence_number() &&
	  token_state == TOKEN_PENDING) {
	token_state= TOKEN_BUSY;
      }
      break;
    case MTPHeader::T_EMPTY_CANCEL:
      if (token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id)) got_token( packet);
      break;
    case MTPHeader::T_EMPTY_HIBERNATE:
      // only master is allowed to send hibernate packet
      // 
      break;
    }    
    break;
  case MTPHeader::T_DATA:
    switch (packet.header.packet_modifier) {
    case MTPHeader::T_DATA_EOM:
      //      cerr << "master:: EOM" << endl;
      if (token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id)) got_token( packet);
      break;
    case MTPHeader::T_DATA_DATA:
    case MTPHeader::T_DATA_EOW:
      if (token_owner == TSAP(packet.addrfrom, 
			      packet.addrfrom_len,
			      packet.header.src_id) &&
	  packet.header.msg_seq_number == web.sequence_number() &&
	  token_state == TOKEN_PENDING) {
	token_state= TOKEN_BUSY;
      }
      break;
    }
    break;
  }
}

int RunMaster::
initialize() {
  (new omni_thread(static_run,this))->start();
  omni_mutex_lock oml(guard);
  while (state != STATE_FAILED &&
	 state != STATE_ESTABLISHED)
    cond_init.wait();
  return (state==STATE_ESTABLISHED);
}

int init_mtp_deamon( PacketMgr &packetmgr) {
  RunMaster *rm= new RunMaster(packetmgr);
  return rm->initialize();
}


#endif // NO_MCAST

//
